#include "tetrisboard.h"

#include "loadgfx.h"
#include <zlib.h>
#include <stdlib.h>

#define TIMEOUT_INTERVAL 25

#define WIDTH 16
#define HEIGHT 32

extern gboolean quitting;

static void tetris_board_class_init (TetrisBoardClass *klass);
static void tetris_board_init       (TetrisBoard      *spriteview);
static void tetris_board_destroy    (GtkObject        *object);
				       
static GsfSpriteViewClass *parent_class = NULL;

enum {
    REMOVED_LINES,
    GAME_OVER,
    NEXT_PIECE,
    RESETTING,
    LAST_SIGNAL
};

static guint tetris_board_signals[LAST_SIGNAL] = { 0 };

GtkType
tetris_board_get_type(void)
{
    static GtkType tetrisboard_type = 0;

    if (!tetrisboard_type) {
	GtkTypeInfo tetrisboard_info = {
	    "TetrisBoard",
	    sizeof(TetrisBoard),
	    sizeof(TetrisBoardClass),
	    (GtkClassInitFunc) tetris_board_class_init,
	    (GtkObjectInitFunc) tetris_board_init,
	    (GtkArgSetFunc) NULL,
	    (GtkArgGetFunc) NULL,
	};

	tetrisboard_type = gtk_type_unique(gsf_spriteview_get_type(),
					   &tetrisboard_info);
    }

    return tetrisboard_type;
}


static void
tetris_board_class_init(TetrisBoardClass *klass)
{
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;

    object_class = (GtkObjectClass*)klass;
    widget_class = (GtkWidgetClass*)klass;
    parent_class = gtk_type_class(gsf_spriteview_get_type());

    tetris_board_signals[REMOVED_LINES] =
	gtk_signal_new("removed_lines",
		       GTK_RUN_FIRST,
		       object_class->type,
		       GTK_SIGNAL_OFFSET(TetrisBoardClass, removed_lines),
		       gtk_marshal_NONE__INT,
		       GTK_TYPE_NONE,
		       1,
		       GTK_TYPE_INT);
    tetris_board_signals[GAME_OVER] =
	gtk_signal_new("game_over",
		       GTK_RUN_FIRST,
		       object_class->type,
		       GTK_SIGNAL_OFFSET(TetrisBoardClass, game_over),
		       gtk_marshal_NONE__NONE,
		       GTK_TYPE_NONE,
		       0);
    tetris_board_signals[NEXT_PIECE] =
	gtk_signal_new("next_piece",
		       GTK_RUN_FIRST,
		       object_class->type,
		       GTK_SIGNAL_OFFSET(TetrisBoardClass, next_piece),
		       gtk_marshal_NONE__POINTER,
		       GTK_TYPE_NONE,
		       1,
		       GTK_TYPE_OBJECT);
    tetris_board_signals[RESETTING] =
	gtk_signal_new("resetting",
		       GTK_RUN_FIRST,
		       object_class->type,
		       GTK_SIGNAL_OFFSET(TetrisBoardClass, resetting),
		       gtk_marshal_NONE__NONE,
		       GTK_TYPE_NONE,
		       0);
    
    gtk_object_class_add_signals(object_class,
				 tetris_board_signals,
				 LAST_SIGNAL);
    
    object_class->destroy = tetris_board_destroy;
    
    klass->removed_lines = NULL;
    klass->game_over = NULL;
    klass->next_piece = NULL;
    klass->resetting = NULL;
}

static gzFile
open_gzfile(const char *file)
{
    gzFile f;

    f = gzopen(file, "r");
    if (!f)
	g_error("Could not read file `%s'", file);

    return f;
}

static void
tetris_board_init(TetrisBoard *tetrisboard)
{
    tetrisboard->width = 0;
    tetrisboard->height = 0;
    tetrisboard->numsprgfx = 0;
    tetrisboard->sprgfx = NULL;

    tetrisboard->block = NULL;
    tetrisboard->board = NULL;
    tetrisboard->next = NULL;
    tetrisboard->piece = NULL;
    tetrisboard->fall = 0;
    tetrisboard->fallspeed = 10;
    tetrisboard->move = 0;
    tetrisboard->movespeed = 2;

    tetrisboard->rcw_pressed = FALSE;
    tetrisboard->rccw_pressed = FALSE;
    tetrisboard->left_pressed = FALSE;
    tetrisboard->right_pressed = FALSE;
    tetrisboard->down_pressed = FALSE;
    tetrisboard->rcw_released = FALSE;
    tetrisboard->rccw_released = FALSE;
    tetrisboard->left_released = FALSE;
    tetrisboard->right_released = FALSE;
    tetrisboard->down_released = FALSE;

    tetrisboard->timer = 0;
}

static void
tetris_board_new_piece(TetrisBoard *tetrisboard)
{
    GsfMovePmSprite *ms;
    GsfPixmapSprite *ps;
    GfxSeq *seq;
    if (!tetrisboard->next) {
	ms = gsf_movepmsprite_new();
	ps = GSF_PIXMAPSPRITE(ms);
	seq = &tetrisboard->sprgfx[rand()%tetrisboard->numsprgfx];
	gsf_pixmapsprite_set(ps,
			     ((tetrisboard->width-1)/2)*10,
			     0,
			     0,
			     seq->seq, seq->num,
			     NULL, 0);
	tetrisboard->next = ms;

	gtk_signal_emit(GTK_OBJECT(tetrisboard),
			tetris_board_signals[NEXT_PIECE],
			tetrisboard->next);
    }
}

GtkWidget *
tetris_board_new(guint width, guint height)
{
    TetrisBoard *tetrisboard;

    gzFile f;
    GdkPixmap *p;
    GsfSpriteGfx *tmp;
    GsfSpriteField *field;
    gint i, j;

    static const char *sprfiles[] = {
	"invl.tar.gz",
	"invs.tar.gz",
	"line.tar.gz",
	"norml.tar.gz",
	"norms.tar.gz",
	"square.tar.gz",
	"tcross.tar.gz"};
    
    tetrisboard = gtk_type_new(tetris_board_get_type());

    tetrisboard->width = width;
    tetrisboard->height = height;
    
    tetrisboard->block = create_filled_gfx(10, 10);    
    gsf_spritegfx_ref(tetrisboard->block);
    gtk_object_sink(GTK_OBJECT(tetrisboard->block));

    field = gsf_spritefield_new(width*10, height*10, 20);
    tmp = create_bordered_gfx(width*10, height*10, 1, 0, 1, 1);
    tetrisboard->board = gsf_pixmapsprite_new();
    gsf_pixmapsprite_set(tetrisboard->board, 0, 0, 0, &tmp, 1, field, -50);
    tetrisboard->piece = NULL;

    f = open_gzfile("gameover.png");
    tmp = loadpng(f);
    gzclose(f);
    tetrisboard->gameover = gsf_pixmapsprite_new();
    gsf_pixmapsprite_set(tetrisboard->gameover,
			 width*5, height*5, 0,
			 &tmp, 1, NULL, 50);
    gsf_sprite_setfield(tetrisboard->gameover, field);
    
    f = open_gzfile("background.png");
    loadpng_pixmap(f, &p, NULL, NULL, NULL);
    gzclose(f);
    
    gsf_spritefield_setbgpixmap(field, p);
    gdk_pixmap_unref(p);

    gsf_spriteview_setfield(GSF_SPRITEVIEW(tetrisboard), field);

    tetrisboard->numsprgfx = sizeof(sprfiles)/sizeof(sprfiles[0]);
    tetrisboard->sprgfx = g_new(GfxSeq, tetrisboard->numsprgfx);
    for (i = 0; i < tetrisboard->numsprgfx; i++) {
	f = open_gzfile(sprfiles[i]);	
	tetrisboard->sprgfx[i].seq=loadpngtar(f, &tetrisboard->sprgfx[i].num);
	for (j = 0; j < tetrisboard->sprgfx[i].num; j++) {
	    gsf_spritegfx_ref(tetrisboard->sprgfx[i].seq[j]);
	    gtk_object_sink(GTK_OBJECT(tetrisboard->sprgfx[i].seq[j]));
	}
	gzclose(f);
    }


    tetris_board_new_piece(tetrisboard);
    return GTK_WIDGET(tetrisboard);
}

static void
tetris_board_destroy(GtkObject *object)
{
    gint i, j;
    GfxSeq *seq;
    TetrisBoard *tetrisboard;

    g_return_if_fail(object != NULL && IS_TETRISBOARD(object));
    tetrisboard = TETRISBOARD(object);

    tetris_board_stop(tetrisboard);

    if (tetrisboard->piece) {
	gsf_sprite_setfield(tetrisboard->piece, NULL);
	tetrisboard->piece = NULL;
    }
    if (tetrisboard->next) {
	gsf_movepmsprite_unref(tetrisboard->next);
	tetrisboard->next = NULL;
    }
    if (tetrisboard->board) {
	gsf_sprite_setfield(tetrisboard->board, NULL);
	tetrisboard->board = NULL;
    }
    if (tetrisboard->gameover) {
	gsf_sprite_setfield(tetrisboard->gameover, NULL);
	tetrisboard->gameover = NULL;
    }
    if (tetrisboard->block) {
	gsf_spritegfx_unref(tetrisboard->block);
	tetrisboard->block = NULL;
    }
    
    if (tetrisboard->sprgfx) {
	for (i = 0; i < tetrisboard->numsprgfx; i++) {
	    seq = &tetrisboard->sprgfx[i];
	    for (j = 0; j < seq->num; j++) {
		gsf_spritegfx_unref(seq->seq[j]);
	    }
	}
	g_free(tetrisboard->sprgfx);
	
	tetrisboard->numsprgfx = 0;
	tetrisboard->sprgfx = NULL;
    }

    if (GTK_OBJECT_CLASS(parent_class)->destroy)
	(* GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}

void
tetris_board_rcw_pressed(TetrisBoard *tb)
{
    tb->rcw_pressed = TRUE;
    tb->rcw_released = FALSE;
}

void
tetris_board_rcw_released(TetrisBoard *tb)
{
    tb->rcw_released = TRUE;
}

void
tetris_board_rccw_pressed(TetrisBoard *tb)
{
    tb->rccw_pressed = TRUE;
    tb->rccw_released = FALSE;
}

void
tetris_board_rccw_released(TetrisBoard *tb)
{
    tb->rccw_released = TRUE;
}

void
tetris_board_left_pressed(TetrisBoard *tb)
{
    tb->left_pressed = TRUE;
    tb->left_released = FALSE;
}

void
tetris_board_left_released(TetrisBoard *tb)
{
    tb->left_released = TRUE;
}

void
tetris_board_right_pressed(TetrisBoard *tb)
{
    tb->right_pressed = TRUE;
    tb->right_released = FALSE;
}

void
tetris_board_right_released(TetrisBoard *tb)
{
    tb->right_released = TRUE;
}

void
tetris_board_down_pressed(TetrisBoard *tb)
{
    tb->down_pressed = TRUE;
    tb->down_released = FALSE;
}

void
tetris_board_down_released(TetrisBoard *tb)
{
    tb->down_released = TRUE;
}

static gboolean
filled_line(GsfSpriteGfx *board, GsfSpriteGfx *block, gint y)
{
    gint boardwidth, blockwidth;
    gint x;

    boardwidth = board->width;
    blockwidth = block->width;

    for (x = 0; x < boardwidth; x += blockwidth) {
	if (!gsf_spritegfx_collides(board, 0, 0, block, x, y))
	    return FALSE;
    }

    return TRUE;
}

static gint
tetris_board_timeout(gpointer data)
{
    static gboolean lock = FALSE;
    gint movex, movey, moveframe;
    TetrisBoard *tb;
    if (!lock) {
	lock = TRUE;
	
	g_return_val_if_fail(data != NULL && IS_TETRISBOARD(data), FALSE);
	if (GTK_OBJECT_DESTROYED(data)) {
	    lock = FALSE;
	    return FALSE;
	}
	
	tb = TETRISBOARD(data);

	if (!tb->piece) {
	    tb->piece = tb->next;
	    tb->next = NULL;
	    tetris_board_new_piece(tb);
	    
	    gsf_sprite_setfield(tb->piece,
				GSF_SPRITEVIEW(tb)->field);
	    gsf_sprite_show(tb->piece);
	    tb->move = tb->movespeed;
	    tb->fall = tb->fallspeed;
	}
	
	if (gsf_sprite_collides(tb->board, tb->piece)) {
	    gtk_signal_emit(GTK_OBJECT(tb),
			    tetris_board_signals[GAME_OVER]);
	    
	    gsf_sprite_show(tb->gameover);
	    gsf_spritefield_update(GSF_SPRITEVIEW(tb)->field);
	    lock = FALSE;
	    return FALSE; /* Game Over */
	}
	
	movex = 0;
	movey = 0;
	moveframe = 0;
	
	if (tb->move-- <= 0) {
	    if (tb->left_pressed)
		movex -= 10;
	    if (tb->right_pressed)
		movex += 10;
	    if (tb->rcw_pressed)
		moveframe += 1;
	    if (tb->rccw_pressed)
		moveframe -= 1;
	}

	if (tb->fall <= 0) { 
	    tb->fall = tb->fallspeed;
	    
	    movey = 10;
	} else
	    tb->fall -= tb->down_pressed ? 10 : 1;

	if (tb->left_released)
	    tb->left_pressed = tb->left_released = FALSE;
	if (tb->right_released)
	    tb->right_pressed = tb->right_released = FALSE;
	if (tb->rcw_released)
	    tb->rcw_pressed = tb->rcw_released = FALSE;
	if (tb->rccw_released)
	    tb->rccw_pressed = tb->rccw_released = FALSE;
	if (tb->down_released)
	    tb->down_pressed = tb->down_released = FALSE;
	
	gsf_movepmsprite_setmove(tb->piece, movex, movey, moveframe);
	if (gsf_sprite_collidesnext(tb->board, tb->piece)) {
	    gsf_movepmsprite_setmove(tb->piece, movex, movey, 0);
	    if (gsf_sprite_collidesnext(tb->board, tb->piece)) {
		movex = 0;
		gsf_movepmsprite_setmove(tb->piece, 0, movey, moveframe);
		if (gsf_sprite_collidesnext(tb->board, tb->piece)) {
		    moveframe = 0;
		    gsf_movepmsprite_setmove(tb->piece, 0, movey, 0);
		}
	    } else
		moveframe = 0;
	}
	
	if (movex || moveframe)
	    tb->move = tb->movespeed;
	    
	if (!gsf_sprite_collidesnext(tb->board, tb->piece)) {
	    gsf_movepmsprite_move(tb->piece);
	} else {
	    gint sy, ey, lines;
	    
	    GsfPixmapSprite *ps = GSF_PIXMAPSPRITE(tb->piece);
	    gsf_pixmapsprite_pastegfx(tb->board,
				      ps,
				      ps->x,
				      ps->y);
	    sy = ps->y - ps->gfx[ps->currentframe]->hoty;
	    ey = sy + ps->gfx[ps->currentframe]->height - 1;
	    
  	    gsf_sprite_setfield(tb->piece, NULL);
	    tb->piece = NULL;

	    lines = 0;
	    while (sy < ey) {
		if (filled_line(tb->board->gfx[tb->board->currentframe],
				tb->block,
				sy)) {
		    GsfRectangle area;
		    area.x = 0;
		    area.y = 0;
		    area.width = tb->width*10;
		    if (sy > 0) {
			area.height = sy;
			gsf_pixmapsprite_pastegfxrect(tb->board,
						      tb->board,
						      0, 10,
						      &area,
						      FALSE);
		    }
		    area.height = 10;
		    gsf_pixmapsprite_clearrect(tb->board, &area);
		    
		    lines++;
		}
		sy += 10;
	    }
	    if (lines)
		gtk_signal_emit(GTK_OBJECT(tb),
				tetris_board_signals[REMOVED_LINES],
				lines);
	    
	}

	gsf_spritefield_update(GSF_SPRITEVIEW(tb)->field);
	
	lock = FALSE;
    }
    
    return TRUE;
}


void
tetris_board_reset(TetrisBoard *tb)
{
    GsfRectangle area;
    tetris_board_stop(tb);

    area.x = 0;
    area.y = 0;
    area.width = tb->width*10;
    area.height = tb->height*10;
    gsf_pixmapsprite_clearrect(tb->board, &area);

    if (tb->piece) {
	gsf_sprite_setfield(tb->piece, NULL);
	tb->piece = NULL;
    }
    if (tb->next) {
	gsf_movepmsprite_unref(tb->next);
	tb->next = NULL;
    }

    gsf_sprite_hide(tb->gameover);
    tetris_board_new_piece(tb);
    gsf_spritefield_update(GSF_SPRITEVIEW(tb)->field);

    gtk_signal_emit(GTK_OBJECT(tb),
		    tetris_board_signals[RESETTING]);
}

void
tetris_board_start(TetrisBoard *tb)
{
    if (tb->timer == 0)
	tb->timer = gtk_timeout_add(TIMEOUT_INTERVAL,
				    tetris_board_timeout,
				    (gpointer)tb);
}

void
tetris_board_stop(TetrisBoard *tb)
{
    if (tb->timer != 0) {
	gtk_timeout_remove(tb->timer);
	tb->timer = 0;
    }
}
