/* GSF - GTK+ SpriteField.  Library to use sprites in GTK+
 * Copyright (C) 1998  Tore Skjellnes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include "gsfbasesprite.h"
#include "gsfspritefield.h"

static void gsf_basesprite_class_init (GsfBaseSpriteClass   *klass);
static void gsf_basesprite_init       (GsfBaseSprite        *basesprite);
static void gsf_basesprite_destroy    (GtkObject            *object);

static void     gsf_basesprite_makevisible   (GsfBaseSprite  *s,
				              gboolean        vis);
static void     gsf_basesprite_draw          (GsfBaseSprite  *s,
				              GsfPainter     *p);
static GSList * gsf_basesprite_approxat_real (GsfBaseSprite  *s);
static GsfCollisionMask *gsf_basesprite_gcm  (GsfBaseSprite  *s,
					      gint           *x,
					      gint           *y);




static GtkObjectClass *parent_class = NULL;

/*
 * Returns a unique type specifier.
 */
GtkType
gsf_basesprite_get_type(void)
{
    static GtkType basesprite_type = 0;

    if (!basesprite_type) {
	GtkTypeInfo basesprite_info = {
	    "GsfBaseSprite",
	    sizeof(GsfBaseSprite),
	    sizeof(GsfBaseSpriteClass),
	    (GtkClassInitFunc) gsf_basesprite_class_init,
	    (GtkObjectInitFunc) gsf_basesprite_init,
	    (GtkArgSetFunc) NULL,
	    (GtkArgGetFunc) NULL,
	};

	basesprite_type = gtk_type_unique(gtk_object_get_type(),
					  &basesprite_info);
    }

    return basesprite_type;
}

/*
 * Initializes the class structure.
 *
 * klass - Structure to initialize
 */
static void
gsf_basesprite_class_init(GsfBaseSpriteClass *klass)
{
    GtkObjectClass *object_class;

    object_class = (GtkObjectClass*)klass;
    parent_class = gtk_type_class(gtk_object_get_type());
    
    object_class->destroy = gsf_basesprite_destroy;
    klass->makevisible = gsf_basesprite_makevisible;
    klass->draw = gsf_basesprite_draw;
    klass->approxat = gsf_basesprite_approxat_real; 
    klass->getcollisionmask = gsf_basesprite_gcm;
    klass->getnextcollisionmask = gsf_basesprite_gcm;
}

/*
 * Initializes an instance
 *
 * basesprite - the instance to initialize
 */
static void
gsf_basesprite_init(GsfBaseSprite *basesprite)
{
    basesprite->z = 0;
    basesprite->visible = FALSE;
    basesprite->field = NULL;
    basesprite->tag = FALSE;

    basesprite->collidecache_clean = FALSE;
    basesprite->collidecache = NULL;
}

/*
 * Creates and returns a new basesprite
 */
GsfBaseSprite *
gsf_basesprite_new(void)
{
    return gtk_type_new(gsf_basesprite_get_type());
}

/*
 * Increases the reference count
 */
void
gsf_basesprite_ref(GsfBaseSprite *basesprite)
{
    g_return_if_fail(basesprite!=NULL && GSF_IS_BASESPRITE(basesprite));

    gtk_object_ref(GTK_OBJECT(basesprite));
}

/*
 * Decreases the reference count, possibly freeing it
 */
void
gsf_basesprite_unref(GsfBaseSprite *basesprite)
{
    g_return_if_fail(basesprite!=NULL && GSF_IS_BASESPRITE(basesprite));

    gtk_object_unref(GTK_OBJECT(basesprite));
}

/*
 * Destroys a sprite
 */
static void
gsf_basesprite_destroy(GtkObject *object)
{
    GsfBaseSprite *basesprite;

    g_return_if_fail(object != NULL && GSF_IS_BASESPRITE(object));

    basesprite = GSF_BASESPRITE(object);

    gsf_basesprite_setfield(basesprite, NULL);

    if (GTK_OBJECT_CLASS(parent_class)->destroy)
	(* GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}

/*
 * Sets the visibility of the sprite, using the virtual function
 * `makevisible'
 *
 * s - sprite
 * visible - TRUE if visible, FALSE if not
 */
void
gsf_basesprite_setvisible(GsfBaseSprite *s, gboolean visible)
{
    g_return_if_fail(s != NULL && GSF_IS_BASESPRITE(s));

    if (s->visible != visible) {
	s->visible = visible;
	GSF_BASESPRITE_CLASS(GTK_OBJECT(s)->klass)->makevisible(s, visible);
    }
}

/*
 * Sets the field member of a sprite.  Dirties the collidecache of the
 * sprite, removes it from the old field and adds it to the new one.
 *
 * s - the sprite
 * f - the new field
 */
void
gsf_basesprite_setfield(GsfBaseSprite *s, GsfSpriteField *f)
{
    g_return_if_fail(s != NULL && GSF_IS_BASESPRITE(s));
    g_return_if_fail(f == NULL || GSF_IS_SPRITEFIELD(f));

    if (f != s->field) {
	gsf_basesprite_dirtycollidecache(s);
	/* Must reference the sprite to keep it alive while changing
           fields.  The spritefield functions `..._removesprite' and
           `..._addsprite' does the actual setting of the field
           member. */
	if (f)
	    gsf_basesprite_ref(s);
	if (s->field)
	    gsf_spritefield_removesprite(s->field, s);
	if (f) {
	    gsf_spritefield_addsprite(f, s);
	    gsf_basesprite_unref(s);
	}
    }
}

/*
 * Sets the z coordinate of a sprite.
 *
 * s - the sprite
 * f - the new field
 */
void
gsf_basesprite_setz(GsfBaseSprite *s, gint z)
{
#ifdef GSF_SPRITELIST_IN_SPRITEFIELD_IS_ORDERED
    GsfSpriteField *f;
#endif
    
    g_return_if_fail(s != NULL && GSF_IS_BASESPRITE(s));

#ifdef GSF_SPRITELIST_IN_SPRITEFIELD_IS_ORDERED
    f = s->field;
    if (f) {
	gsf_basesprite_ref(s);
	gsf_spritefield_removesprite(f, s);
    }
#endif
    
    s->z = z;
    
#ifdef GSF_SPRITELIST_IN_SPRITEFIELD_IS_ORDERED
    if (f) {
	gsf_spritefield_addsprite(f, s);
	gsf_basesprite_unref(s);
    }
#endif
}


/*
 * Calls the virtual function `approxat' of a sprite and returns a list of
 * sprites that maybe collides with the supplied sprite.
 *
 * s - the sprite
 */
GSList *
gsf_basesprite_approxat(GsfBaseSprite *s)
{
    GtkObject *obj = GTK_OBJECT(s);
    return (* GSF_BASESPRITE_CLASS(obj->klass)->approxat)(s);
}

/*
 * Returns the sprites colliding with the supplied sprite.  Updates
 * the collide_cache if needed.
 *
 * s - the sprite
 */
GSList *
gsf_basesprite_at(GsfBaseSprite *s)
{
    GSList *list;
    GSList *iter;
    
    g_return_val_if_fail(s != NULL && GSF_IS_BASESPRITE(s), NULL);
    
    if (!s->collidecache_clean) {
	g_slist_free(s->collidecache);
	s->collidecache = NULL;

	list = gsf_basesprite_approxat(s);
	iter = list;
	while (iter) {
	    if (s != iter->data && gsf_basesprite_collides(s, iter->data))
		s->collidecache = g_slist_prepend(s->collidecache, iter->data);
	    iter = iter->next;
	}
	g_slist_free(list);
	s->collidecache_clean = TRUE;
    }
    return s->collidecache;
}

/*
 * Returns TRUE if the point intersects with the supplied sprite,
 * FALSE else.
 *
 * s - the sprite
 *
 * x - x coordinate
 *
 * y - y coordinate
 */
gboolean
gsf_basesprite_spriteat(GsfBaseSprite *s, gint x, gint y)
{
    GtkObject *obj;
    GsfBaseSpriteClass *cl;
    GsfCollisionMask *mask;
    gint mx, my;
    
    obj = GTK_OBJECT(s);
    cl = GSF_BASESPRITE_CLASS(obj->klass);
    mask = (* cl->getcollisionmask)(s, &mx, &my);
    if (!mask)
	return FALSE;

    return gsf_collisionmask_collides_at(mask, x - mx, y - my);
}

/*
 * Returns TRUE if the two sprites collides.  Does not use the
 * collision_cache.
 *
 * s - first sprite
 *
 * o - second sprite
 */
gboolean
gsf_basesprite_collides(GsfBaseSprite *s, GsfBaseSprite *o)
{   
    GtkObject *obj;
    GsfBaseSpriteClass *cl;
    GsfCollisionMask *smask;
    gint sx, sy;
    GsfCollisionMask *omask;
    gint ox, oy;
    
    obj = GTK_OBJECT(s);
    cl = GSF_BASESPRITE_CLASS(obj->klass);
    smask = (* cl->getcollisionmask)(s, &sx, &sy);

    if (!smask)
	return FALSE;
    
    obj = GTK_OBJECT(o);
    cl = GSF_BASESPRITE_CLASS(obj->klass);
    omask = (* cl->getcollisionmask)(o, &ox, &oy);

    if (!omask)
	return FALSE;

    return gsf_collisionmask_collides(smask, sx, sy,
				      omask, ox, oy);
}

/*
 * Returns TRUE if the two sprites collides after "moving".  Does not
 * use the collision_cache.
 *
 * s - first sprite
 *
 * o - second sprite */
gboolean
gsf_basesprite_collidesnext(GsfBaseSprite *s, GsfBaseSprite *o)
{
    GtkObject *obj;
    GsfBaseSpriteClass *cl;
    GsfCollisionMask *smask;
    gint sx, sy;
    GsfCollisionMask *omask;
    gint ox, oy;
    
    obj = GTK_OBJECT(s);
    cl = GSF_BASESPRITE_CLASS(obj->klass);
    smask = (* cl->getnextcollisionmask)(s, &sx, &sy);

    if (!smask)
	return FALSE;
    
    obj = GTK_OBJECT(o);
    cl = GSF_BASESPRITE_CLASS(obj->klass);
    omask = (* cl->getnextcollisionmask)(o, &ox, &oy);

    if (!omask)
	return FALSE;

    return gsf_collisionmask_collides(smask, sx, sy,
				      omask, ox, oy);
}

/*
 * Dirties the collision cache of the supplied sprite, and all sprites
 * in the collision cache (recursive). 
 *
 * s - the sprite
 */
void
gsf_basesprite_dirtycollidecache(GsfBaseSprite *s)
{
    if (s->collidecache_clean) {
	GSList *iter;
	s->collidecache_clean = FALSE;

	iter = s->collidecache;
	while (iter) {
	    gsf_basesprite_dirtycollidecache(GSF_BASESPRITE(iter->data));
	    iter = iter->next;
	}
	g_slist_free(s->collidecache);
	s->collidecache = NULL;
    }
}


/*
 * Default `approxat' function.  Returns NULL.
 */
static GSList *
gsf_basesprite_approxat_real(GsfBaseSprite *s)
{
    return NULL;
}

/*
 * Default `makevisible' function.  Writes a warning message.
 */
static void
gsf_basesprite_makevisible(GsfBaseSprite *s, gboolean vis)
{
    g_warning("GsfBaseSprite: Call to abstract function `makevisible'");
}

/*
 * Default `draw' function.  Writes a warning message.
 */
static void
gsf_basesprite_draw(GsfBaseSprite *s, GsfPainter *p)
{
    g_warning("GsfBaseSprite: Call to abstract function `draw'");
}


/*
 * Default 'get_collisionmask' function.  Returns NULL.
 */
static GsfCollisionMask *
gsf_basesprite_gcm(GsfBaseSprite *s,
		   gint          *x,
		   gint          *y)
{
    return NULL;
}




