/* GSF - GTK+ SpriteField.  Library to use sprites in GTK+
 * Copyright (C) 1998  Tore Skjellnes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
/*
 * BaseSprite
 *
 * $Id: gsfbasesprite.h,v 1.10 1998/07/30 03:35:40 torsk Exp $
 *
 */
#ifndef __GSF_BASESPRITE_H__
#define __GSF_BASESPRITE_H__ 1

#include <gtk/gtk.h>
#include <glib.h>

#include "gsfpainter.h"
#include "gsfcollisionmask.h"

/* FIX: Commented out due to annoing Emacs indent:
#ifdef __cplusplus
extern "C" {
#endif
*/

#define GSF_BASESPRITE(obj)             (GTK_CHECK_CAST (obj, gsf_basesprite_get_type (), GsfBaseSprite))
#define GSF_BASESPRITE_CLASS(klass)     (GTK_CHECK_CLASS_CAST (klass, gsf_basesprite_get_type (), GsfBaseSpriteClass))
#define GSF_IS_BASESPRITE(obj)          (GTK_CHECK_TYPE (obj, gsf_basesprite_get_type ()))

#include "gsftypedefs.h"

/* GsfBaseSprite is the instance of the base class of the sprite
 * hierarcy.
 *
 * Members:
 *
 *  object - parent data
 *
 *  z - the z coordinate of this sprite, used when deciding the
 *  drawing * order.
 *
 *  visible - a boolean indicating if the sprite is to be drawn.
 *
 *  field - the field this sprite belongs to
 *
 *  tag - a temporary used by spritefield to speed up adding sprites
 *  to a list, with only one of each sprite.  Should always be FALSE.
 *  See * the static function `gsf_spritefield_at_rect_unchecked'.
 *
 *  collidecache_clean - TRUE if the collidecache is valid.
 *
 *  collidecache - a slist of all the sprites colliding with this
 *  sprite if `collidecache_clean' is TRUE.
 */
struct _GsfBaseSprite
{
    GtkObject object;

    gint z;
    gboolean visible;
    GsfSpriteField *field;

    gboolean tag;

    gboolean collidecache_clean;
    GSList *collidecache;
};

/*
 * GsfBaseSpriteClass is the base class of the sprite hierarcy.
 *
 * Members:
 *
 *  parent_class - the parent class.
 *
 *  makevisible - makes a sprite visible/invisible.  Default is to
 *  write a warning message (abstract function).
 *
 *  draw - draws a sprite.  Default is to write a warning message
 *  (abstract function).
 *
 *  approxat - returns a list of sprites that maybe collides with the
 *  supplied sprite.  Default is to return NULL;
 *
 *  getcollisionmask - returns the collision mask and position, or
 *  NULL if collision detection is not wanted for this sprite
 *  (default).
 *
 *  getcollisionmasknext - returns the collision mask and position
 *  after some operation (for instance a movement), or NULL if
 *  collision detection is not wanted for this sprite (default).
 */
struct _GsfBaseSpriteClass
{
    GtkObjectClass parent_class;

    void               (* makevisible)         (GsfBaseSprite *s,
						gboolean vis);
    void               (* draw)                (GsfBaseSprite *s,
						GsfPainter *p);    
    GSList *           (* approxat)            (GsfBaseSprite *s);
    GsfCollisionMask * (* getcollisionmask)    (GsfBaseSprite *s,
						gint *x,
						gint *y);
    GsfCollisionMask * (* getnextcollisionmask)(GsfBaseSprite *s,
						gint *x,
						gint *y);
};

GtkType           gsf_basesprite_get_type         (void);
GsfBaseSprite*    gsf_basesprite_new              (void);
void              gsf_basesprite_ref              (GsfBaseSprite *sprite);
void              gsf_basesprite_unref            (GsfBaseSprite *sprite);
void              gsf_basesprite_setvisible       (GsfBaseSprite *s,
					           gboolean visible);
void              gsf_basesprite_setfield         (GsfBaseSprite *s,
					           GsfSpriteField *f);
void              gsf_basesprite_setz             (GsfBaseSprite *s,
					           gint z);
GSList *          gsf_basesprite_at               (GsfBaseSprite *s);
GSList *          gsf_basesprite_approxat         (GsfBaseSprite *s);
gboolean          gsf_basesprite_spriteat         (GsfBaseSprite *s,
					           gint x,
					           gint y);
gboolean          gsf_basesprite_collides         (GsfBaseSprite *s,
					           GsfBaseSprite *o);
gboolean          gsf_basesprite_collidesnext     (GsfBaseSprite *s,
					           GsfBaseSprite *o);
void              gsf_basesprite_dirtycollidecache(GsfBaseSprite *s);

#define gsf_basesprite_show(s) gsf_basesprite_setvisible((s), TRUE)
#define gsf_basesprite_hide(s) gsf_basesprite_setvisible((s), FALSE)

/* Convenience macros: */
#define gsf_sprite_setvisible(s, v) \
            gsf_basesprite_setvisible  (GSF_BASESPRITE(s), (v))
#define gsf_sprite_show(s) \
            gsf_basesprite_show        (GSF_BASESPRITE(s))
#define gsf_sprite_hide(s) \
            gsf_basesprite_hide        (GSF_BASESPRITE(s))
#define gsf_sprite_isvisible(s) \
            (GSF_BASESPRITE(s)->visible)
#define gsf_sprite_setfield(s, f) \
            gsf_basesprite_setfield    (GSF_BASESPRITE(s), (f))
#define gsf_sprite_setz(s, z) \
            gsf_basesprite_setz        (GSF_BASESPRITE(s), (z))
#define gsf_sprite_spriteat(s, x, y) \
            gsf_basesprite_spriteat    (GSF_BASESPRITE(s), (x), (y))
#define gsf_sprite_approxat(s) \
            gsf_basesprite_approxat    (GSF_BASESPRITE(s))
#define gsf_sprite_at(s) \
            gsf_basesprite_at          (GSF_BASESPRITE(s))
#define gsf_sprite_collides(s, o) \
            gsf_basesprite_collides    (GSF_BASESPRITE(s), GSF_BASESPRITE(o))
#define gsf_sprite_collidesnext(s, o) \
            gsf_basesprite_collidesnext(GSF_BASESPRITE(s), GSF_BASESPRITE(o))

/* FIX: Commented out due to annoing Emacs indent:
#ifdef __cplusplus
}
#endif
*/  

#endif /* __GSF_BASESPRITE_H__ */
