/* GSF - GTK+ SpriteField.  Library to use sprites in GTK+
 * Copyright (C) 1998  Tore Skjellnes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include "gsfmemarr.h"

#define SIZE_ALIGN       sizeof(double)
#define START_CHUNK_SIZE 16
#define MAX_CHUNK_SIZE   8096

static void
adjust_element_size(GsfMemArr *ma)
{
    guint element_size = ma->element_size;
    
    /* Make element_size aligned with double */
    element_size = ((element_size + SIZE_ALIGN - 1)/SIZE_ALIGN) * SIZE_ALIGN;
    
    /* Make sure element_size is at least as big as a pointer*/
    if (element_size < sizeof(gpointer))
	element_size = sizeof(gpointer);

    ma->element_size = element_size;
}

void
gsf_memarr_init(GsfMemArr *ma, guint element_size)
{
    ma->next_free = NULL;
    ma->element_size = element_size;
    ma->chunk_list = NULL;
    ma->next_chunk_size = 0;
}

GsfMemArr*
gsf_memarr_create(guint element_size)
{
    GsfMemArr *ma;

    ma = g_new(GsfMemArr, 1);
    gsf_memarr_init(ma, element_size);

    return ma;
}

void
gsf_memarr_destroy(GsfMemArr *ma)
{
    GSList *iter = ma->chunk_list;
    while (iter) {
	g_free(iter->data);
	iter = iter->next;
    }
    g_slist_free(ma->chunk_list);
    ma->chunk_list = NULL;
    
    ma->next_free = NULL;
    ma->next_chunk_size = START_CHUNK_SIZE;
}

void
gsf_memarr_addchunk(GsfMemArr *ma)
{
    gpointer chunk;
    gpointer p;
    gpointer e;

    /* Check chunk size (0 if initializing) */
    if (ma->next_chunk_size < START_CHUNK_SIZE) {
	ma->next_chunk_size = START_CHUNK_SIZE;
	adjust_element_size(ma);
    }
    
    /* Allocate chunk */
    chunk = g_malloc(ma->element_size * ma->next_chunk_size);
    ma->chunk_list = g_slist_prepend(ma->chunk_list, chunk);

    /* Initialize chunk */
    p = chunk;
    e = (gpointer)((guchar *)p + ma->element_size * (ma->next_chunk_size - 1));
    while (p < e) {
	gpointer n;

	n = (gpointer)((guchar *)p + ma->element_size);
	*((gpointer *)p) = n;
	p = n;
    }
    *((gpointer *)p) = ma->next_free;
    ma->next_free = chunk;
    
    /* Increase next chunk size */
    if (ma->next_chunk_size < MAX_CHUNK_SIZE) {
	ma->next_chunk_size = ma->next_chunk_size << 1;
	if (ma->next_chunk_size > MAX_CHUNK_SIZE)
	    ma->next_chunk_size = MAX_CHUNK_SIZE;
    }
}

gpointer
gsf_memarr_alloc(GsfMemArr *ma)
{
    gpointer p;
    if (!ma->next_free)
	gsf_memarr_addchunk(ma);

    p = ma->next_free;
    ma->next_free = *((gpointer *)p);
    return p;
}

void
gsf_memarr_free(GsfMemArr *ma, gpointer p)
{
    *((gpointer *)p) = ma->next_free;
    ma->next_free = p;
}

