/* GSF - GTK+ SpriteField.  Library to use sprites in GTK+
 * Copyright (C) 1998  Tore Skjellnes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include "gsfmovepmsprite.h"
#include "gsfspritefield.h"

static void gsf_movepmsprite_class_init    (GsfMovePmSpriteClass *klass);
static void gsf_movepmsprite_init          (GsfMovePmSprite      *sprite);
static GsfCollisionMask *gsf_movepmsprite_gncm(GsfBaseSprite     *s,
					       gint              *x,
					       gint              *y);

static GsfPixmapSpriteClass *parent_class = NULL;

GtkType
gsf_movepmsprite_get_type(void)
{
    static GtkType sprite_type = 0;

    if (!sprite_type) {
	GtkTypeInfo gsf_movepmsprite_info = {
	    "GsfMovePmSprite",
	    sizeof(GsfMovePmSprite),
	    sizeof(GsfMovePmSpriteClass),
	    (GtkClassInitFunc) gsf_movepmsprite_class_init,
	    (GtkObjectInitFunc) gsf_movepmsprite_init,
	    (GtkArgSetFunc) NULL,
	    (GtkArgGetFunc) NULL,
	};

	sprite_type = gtk_type_unique(gsf_pixmapsprite_get_type(),
				      &gsf_movepmsprite_info);
    }

    return sprite_type;
}

static void
gsf_movepmsprite_class_init(GsfMovePmSpriteClass *klass)
{
    GtkObjectClass *object_class;
    GsfPixmapSpriteClass *pixmapsprite_class;
    GsfBaseSpriteClass *basesprite_class;

    object_class = (GtkObjectClass*)klass;
    pixmapsprite_class = (GsfPixmapSpriteClass *)klass;
    basesprite_class = (GsfBaseSpriteClass *)klass;
    
    parent_class = gtk_type_class(gsf_pixmapsprite_get_type());

    basesprite_class->getnextcollisionmask = gsf_movepmsprite_gncm;
}

static void
gsf_movepmsprite_init(GsfMovePmSprite *sprite)
{
    sprite->deltax = 0;
    sprite->deltay = 0;
    sprite->deltaframe = 0;
    sprite->minx = G_MININT;
    sprite->miny = G_MININT;
    sprite->maxx = G_MAXINT;
    sprite->maxy = G_MAXINT;
    sprite->action = GSF_MOVEPMSPRITE_NONE;
}

GsfMovePmSprite *
gsf_movepmsprite_new(void)
{
    return gtk_type_new(gsf_movepmsprite_get_type());
}

void
gsf_movepmsprite_ref(GsfMovePmSprite *movepmsprite)
{
    g_return_if_fail(movepmsprite!=NULL && GSF_IS_MOVEPMSPRITE(movepmsprite));

    gtk_object_ref(GTK_OBJECT(movepmsprite));
}

void
gsf_movepmsprite_unref(GsfMovePmSprite *movepmsprite)
{
    g_return_if_fail(movepmsprite!=NULL && GSF_IS_MOVEPMSPRITE(movepmsprite));

    gtk_object_unref(GTK_OBJECT(movepmsprite));
}

void
gsf_movepmsprite_setmove(GsfMovePmSprite *sprite,
			 gint deltax,
			 gint deltay,
			 gint deltaframe)
{
    g_return_if_fail(sprite != NULL && GSF_IS_MOVEPMSPRITE(sprite));

    sprite->deltax = deltax;
    sprite->deltay = deltay;
    sprite->deltaframe = deltaframe;
}

void
gsf_movepmsprite_setbounds(GsfMovePmSprite *sprite,
			   gint minx,
			   gint miny,
			   gint maxx,
			   gint maxy,
			   GsfMovePmAction action)
{
    g_return_if_fail(sprite != NULL && GSF_IS_MOVEPMSPRITE(sprite));

    sprite->minx = minx;
    sprite->miny = miny;
    sprite->maxx = maxx;
    sprite->maxy = maxy;
    sprite->action = action;
}


static gint
gsf_movepmsprite_domove(GsfMovePmSprite *sprite,
			gint *newx, gint *newy, gint *newframe)
{
    GsfPixmapSprite *pixmapsprite;
    gint ret;
    
    pixmapsprite = GSF_PIXMAPSPRITE(sprite);
    ret = 0;
    
    *newx = pixmapsprite->x + sprite->deltax;
    *newy = pixmapsprite->y + sprite->deltay;
    *newframe = pixmapsprite->currentframe + sprite->deltaframe;

    if (sprite->action == GSF_MOVEPMSPRITE_BOUNCE) {
	/* Doesn't bounce right if very small space compared to delta */
	if (*newx < sprite->minx || *newx > sprite->maxx) {
	    gint tmp = *newx < sprite->minx ? sprite->minx : sprite->maxx;
	    *newx = 2*tmp - *newx;

	    ret |= 1;
	}
	if (*newy < sprite->miny || *newy > sprite->maxy) {
	    gint tmp = *newy < sprite->miny ? sprite->miny : sprite->maxy;
	    *newy = 2*tmp - *newy;

	    ret |= 2;
	}
    } else if (sprite->action == GSF_MOVEPMSPRITE_STOP) {
	gint dx = sprite->deltax;
	gint dy = sprite->deltay;
	if (*newx < sprite->minx || *newx > sprite->maxx) {
	    gint tmp = *newx < sprite->minx ? sprite->minx : sprite->maxx;
	    
	    *newx = tmp;
	    *newy = dy*(tmp - *newx)/dx;

	    ret |= 4;
	}
	if (*newy < sprite->miny || *newy > sprite->maxy) {
	    gint tmp = *newy < sprite->miny ? sprite->miny : sprite->maxy;
	    
	    *newx = dx*(tmp - *newy)/dy;
	    *newy = tmp;

	    ret |= 4;
	}
    }

    if ((guint)*newframe >= (guint)pixmapsprite->numframes) {
	*newframe %= pixmapsprite->numframes;
	if (*newframe < 0)
	    *newframe += pixmapsprite->numframes;
    }

    return ret;
}

void
gsf_movepmsprite_move(GsfMovePmSprite *sprite)
{
    gint newx, newy, newframe, ret;

    g_return_if_fail(sprite != NULL && GSF_IS_MOVEPMSPRITE(sprite));

    ret = gsf_movepmsprite_domove(sprite, &newx, &newy, &newframe);
    if (ret & 1)
	sprite->deltax = -sprite->deltax;
    if (ret & 2)
	sprite->deltay = -sprite->deltay;
    if (ret & 4) {
	sprite->deltax = 0;
	sprite->deltay = 0;
    }
    
    gsf_pixmapsprite_setxyf(GSF_PIXMAPSPRITE(sprite), newx, newy, newframe);
}

static GsfCollisionMask *
gsf_movepmsprite_gncm(GsfBaseSprite  *s,
		      gint           *x,
		      gint           *y)
{
    gint newx, newy, newframe;
    GsfPixmapSprite *ps;
    GsfMovePmSprite *ms;
    GsfSpriteGfx *tmp;
    g_return_val_if_fail(s != NULL && GSF_IS_MOVEPMSPRITE(s), NULL);

    ms = GSF_MOVEPMSPRITE(s);
    ps = GSF_PIXMAPSPRITE(s);

    newx = ps->x;
    newy = ps->y;
    newframe = ps->currentframe;
    gsf_movepmsprite_domove(ms, &newx, &newy, &newframe);

    if (!ps->gfx)
	return NULL;
    
    tmp = ps->gfx[newframe];
    if (x)
	*x = newx - tmp->hotx;
    if (y)
	*y = newy - tmp->hoty;
    return tmp->collmask;
}

