/* GSF - GTK+ SpriteField.  Library to use sprites in GTK+
 * Copyright (C) 1998  Tore Skjellnes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include "gsfpainter.h"

static void gsf_painter_class_init (GsfPainterClass *klass);
static void gsf_painter_init       (GsfPainter      *painter);
static void gsf_painter_destroy    (GtkObject       *object);

static GtkObjectClass *parent_class = NULL;

GtkType
gsf_painter_get_type(void)
{
    static GtkType painter_type = 0;

    if (!painter_type) {
	GtkTypeInfo painter_info = {
	    "GsfPainter",
	    sizeof(GsfPainter),
	    sizeof(GsfPainterClass),
	    (GtkClassInitFunc) gsf_painter_class_init,
	    (GtkObjectInitFunc) gsf_painter_init,
	    (GtkArgSetFunc) NULL,
	    (GtkArgGetFunc) NULL,
	};

	painter_type = gtk_type_unique(gtk_object_get_type(),
				       &painter_info);
    }

    return painter_type;
}

static void
gsf_painter_class_init(GsfPainterClass *klass)
{
    GtkObjectClass *object_class;

    object_class = (GtkObjectClass*)klass;
    parent_class = gtk_type_class(gtk_object_get_type());
    
    object_class->destroy = gsf_painter_destroy;
}

static void
gsf_painter_init(GsfPainter *painter)
{
    painter->drawable = NULL;
    painter->colormap = NULL;
    painter->gc = NULL;
    painter->originx = 0;
    painter->originy = 0;
}

GsfPainter *
gsf_painter_new(void)
{
    return gtk_type_new(gsf_painter_get_type());
}

void
gsf_painter_ref(GsfPainter *painter)
{
    g_return_if_fail(painter != NULL && GSF_IS_PAINTER(painter));
    gtk_object_ref(GTK_OBJECT(painter));
}

void
gsf_painter_unref(GsfPainter *painter)
{
    g_return_if_fail(painter != NULL && GSF_IS_PAINTER(painter));
    gtk_object_unref(GTK_OBJECT(painter));
}

static void
gsf_painter_destroy(GtkObject *object)
{
    GsfPainter *painter;

    g_return_if_fail(object != NULL && GSF_IS_PAINTER(object));

    painter = GSF_PAINTER(object);
    gsf_painter_set_drawable(painter, NULL, NULL, NULL);

    if (GTK_OBJECT_CLASS(parent_class)->destroy)
	(* GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}

void
gsf_painter_set_drawable(GsfPainter *painter,
			 GdkDrawable *drawable,
			 GdkColormap *colormap,
			 GdkGC *gc)
{
    g_return_if_fail(painter != NULL && GSF_IS_PAINTER(painter));

    if (painter->drawable) gdk_pixmap_unref(painter->drawable);
    if (painter->colormap) gdk_colormap_unref(painter->colormap);
    if (painter->gc)       gdk_gc_unref(painter->gc);

    painter->drawable = drawable ? gdk_pixmap_ref(drawable)   : NULL; /*FIX?*/
    painter->colormap = colormap ? gdk_colormap_ref(colormap) : NULL;
    painter->gc       = gc       ? gdk_gc_ref(gc)             : NULL;
}

void
gsf_painter_init_pixmap(GsfPainter *painter,
			GtkWidget *widget,
			guint width,
			guint height)
{
    GdkPixmap *pixmap;
    GdkColormap *colormap;
    GdkGC *gc;
    
    g_return_if_fail(widget != NULL && GTK_IS_WIDGET(widget));

    pixmap = gdk_pixmap_new(widget->window,
			    width,
			    height,
			    -1);
    gc = gdk_gc_new(widget->window);
    colormap = gdk_window_get_colormap(widget->window);
    
    gsf_painter_set_drawable(painter, pixmap, colormap, gc);

    gdk_pixmap_unref(pixmap);
    gdk_gc_unref(gc);
    /*gdk_colormap_unref(colormap);*/
}

void
gsf_painter_set_origin(GsfPainter *p,
		       gint originx,
		       gint originy)
{
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));

    p->originx = originx;
    p->originy = originy;
}

void
gsf_painter_get_origin(GsfPainter *p,
		       gint *originx,
		       gint *originy)
{
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));

    *originx = p->originx;
    *originy = p->originy;
}

void
gsf_painter_alloc_gdkcolor(GsfPainter *p,
			   GdkColor *c)
{
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));
    gdk_color_alloc(p->colormap, c);
}

void
gsf_painter_alloc_rgbcolor(GsfPainter *p,
			   guint16 r,
			   guint16 g,
			   guint16 b,
			   GdkColor *c)
{
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));
    g_return_if_fail(c != NULL);

    c->red = r;
    c->green = g;
    c->blue = b;
    gdk_color_alloc(p->colormap, c);
}

void
gsf_painter_set_foreground(GsfPainter *p,
			   GdkColor *c)
{
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));
    gdk_gc_set_foreground(p->gc, c);
}

void
gsf_painter_set_background(GsfPainter *p,
			   GdkColor *c)
{
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));
    gdk_gc_set_background(p->gc, c);
}

void
gsf_painter_fill_rectangle(GsfPainter *p,
			   GsfRectangle *area)
{
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));
    g_return_if_fail(area != NULL);
    
    gdk_draw_rectangle(p->drawable, p->gc, TRUE,
		       area->x + p->originx, area->y + p->originy,
		       area->width, area->height);
}

/* static gboolean */
/* gsf_painter_clip(GsfPainter *p, */
/* 		 gint *x, gint *y, guint *w, guint *h) */
/* { */
/*     gint pwidth, pheight; */
/*     gdk_window_get_size(p->drawable, &pwidth, &pheight); */
/*     if ((guint)*x >= (guint)pwidth) { */
/* 	if (*x >= pwidth) */
/* 	    return FALSE; */
 	/* x is < 0: */ 
/* 	*w += *x; */
/* 	if ((gint)*w <= 0) */
/* 	    return FALSE; */
/* 	*x = 0; */
/*     } */
/*     if (*w > pwidth) */
/* 	*w = pwidth; */
    
/*     if ((guint)*y >= (guint)pheight) { */
/* 	if (*y >= pheight) */
/* 	    return FALSE; */
 	/* y is < 0: */
/* 	*h += *y; */
/* 	if ((gint)*h <= 0) */
/* 	    return FALSE; */
/* 	*y = 0; */
/*     } */
/*     if (*h > pheight) */
/* 	*h = pheight; */
    
/*     return TRUE; */
/* } */

void
gsf_painter_draw_tiled_pixmap(GsfPainter *p,
			      GdkPixmap *pixmap,
			      gint offsetx, gint offsety,
			      GsfRectangle *area)
{
    GsfRectangle r;
    gint pwidth, pheight;
    gint sx, ex, sy, ey, cx, cy;
    
    /* Could maybe use window background stuff?? */
    
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));

    gdk_window_get_size(pixmap, &pwidth, &pheight);
    if (pwidth == 0 || pheight == 0)
	return;

    r.x = -p->originx;
    r.y = -p->originy;
    gdk_window_get_size(p->drawable, &r.width, &r.height);
    gsf_rectangle_intersection(&r, &r, area);

    if (gsf_rectangle_isempty(&r))
	return;
    
    r.x += p->originx;
    r.y += p->originy;

    offsetx += p->originx;
    if (offsetx < 0)
	offsetx = pwidth - offsetx;
    offsetx %= pwidth;

    offsety += p->originy;
    if (offsety < 0)
	offsety = pheight - offsety;
    offsety %= pheight;


    /* FIX: check these values (may be too conservative) */
    sx = (r.x/pwidth)*pwidth - offsetx;
    ex = sx + ((r.width + offsetx)/pwidth)*pwidth;
    sy = (r.y/pheight)*pheight - offsety;
    ey = sy + ((r.height + offsety)/pheight)*pheight;
    for (cy = sy; cy <= ey; cy += pheight) {
	gint rsy = 0;
	gint rcy = cy;
	gint rh = pheight;
	if (r.y - rcy > 0) {
	    rsy = r.y - rcy;
	    rh -= r.y - rcy;
	    if (rh <= 0) continue;
	    rcy = r.y;
	}
	if (rcy + rh > r.height) {
	    rh = r.height - rcy;
	    if (rh <= 0) continue;
	}
	for (cx = sx; cx <= ex; cx += pwidth) {
	    gint rsx = 0;
	    gint rcx = cx;
	    gint rw = pwidth;
	    
	    if (r.x - rcx > 0) {
		rsx = r.x - rcx;
		rw -= r.x - rcx;
		if (rw <= 0) continue;
		rcx = r.x;
	    }
	    if (rcx + rw > r.width) {
		rw = r.width - rcx;
		if (rw <= 0) continue;
	    }

	    gdk_draw_pixmap(p->drawable, p->gc, pixmap,
			    rsx, rsy,
			    rcx, rcy,
			    rw, rh);
	}
    }
}

void
gsf_painter_draw_pixmap(GsfPainter *p,
			GdkPixmap *pixmap,
			GsfRectangle *area)
{
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));
    
    gdk_draw_pixmap(p->drawable, p->gc, pixmap,
		    0, 0,
		    area->x + p->originx, area->y + p->originy,
		    area->width, area->height);
}

void
gsf_painter_draw_masked_pixmap(GsfPainter *p,
			       GdkPixmap *pixmap,
			       GdkBitmap *mask,
			       GsfRectangle *area)
{
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));
    
    gdk_gc_set_clip_mask(p->gc, mask);
    gdk_gc_set_clip_origin(p->gc, area->x + p->originx, area->y + p->originy);

    gdk_draw_pixmap(p->drawable, p->gc, pixmap,
		    0, 0,
		    area->x + p->originx, area->y + p->originy,
		    area->width, area->height);
    
    gdk_gc_set_clip_mask(p->gc, NULL);
}

void
gsf_painter_draw_masked_tiled_pixmap(GsfPainter *p,
				     GdkPixmap *pixmap,
				     GdkBitmap *mask,
				     gint offsetx, gint offsety,
				     GsfRectangle *area)
{
    GsfRectangle r;
    gint pwidth, pheight;
    gint sx, ex, sy, ey, cx, cy;
    
    /* Could maybe use window background stuff?? */
    
    g_return_if_fail(p != NULL && GSF_IS_PAINTER(p));

    gdk_window_get_size(pixmap, &pwidth, &pheight);
    if (pwidth == 0 || pheight == 0)
	return;

    r.x = -p->originx;
    r.y = -p->originy;
    r.width = pwidth;
    r.height = pheight;
    gsf_rectangle_intersection(&r, &r, area);

    if (gsf_rectangle_isempty(&r))
	return;
    
    r.x += p->originx;
    r.y += p->originy;

    offsetx += p->originx;
    if (offsetx < 0)
	offsetx = pwidth - offsetx;
    offsetx %= pwidth;

    offsety += p->originy;
    if (offsety < 0)
	offsety = pheight - offsety;
    offsety %= pheight;

    
    
    gdk_gc_set_clip_mask(p->gc, mask);
    sx = (r.x/pwidth)*pwidth - offsetx;
    ex = sx + ((r.width + offsetx)/pwidth)*pwidth;
    sy = (r.y/pheight)*pheight - offsety;
    ey = sy + ((r.height + offsety)/pheight)*pheight;
    for (cy = sy; cy <= ey; cy += pheight) {
	gint rsy = 0;
	gint rcy = cy;
	gint rh = pheight;
	if (r.y - rcy > 0) {
	    rsy = r.y - rcy;
	    rh -= r.y - rcy;
	    if (rh <= 0) continue;
	    rcy = r.y;
	}
	if (rcy + rh > r.height) {
	    rh = r.height - rcy;
	    if (rh <= 0) continue;
	}
	for (cx = sx; cx <= ex; cx += pwidth) {
	    gint rsx = 0;
	    gint rcx = cx;
	    gint rw = pwidth;
	    
	    if (r.x - rcx > 0) {
		rsx = r.x - rcx;
		rw -= r.x - rcx;
		if (rw <= 0) continue;
		rcx = r.x;
	    }
	    if (rcx + rw > r.width) {
		rw = r.width - rcx;
		if (rw <= 0) continue;
	    }
	    
	    gdk_gc_set_clip_origin(p->gc, cx, cy);

	    gdk_draw_pixmap(p->drawable, p->gc, pixmap,
			    rsx, rsy,
			    rcx, rcy,
			    rw, rh);
	}
    }
    gdk_gc_set_clip_mask(p->gc, NULL);
}



