/* GSF - GTK+ SpriteField.  Library to use sprites in GTK+
 * Copyright (C) 1998  Tore Skjellnes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include "gsfpixmapsprite.h"
#include "gsfspritefield.h"

static void gsf_pixmapsprite_class_init   (GsfPixmapSpriteClass *klass);
static void gsf_pixmapsprite_init         (GsfPixmapSprite      *sprite);
static void gsf_pixmapsprite_destroy      (GtkObject            *object);
static void gsf_pixmapsprite_makevisible  (GsfBaseSprite *s,
					   gboolean vis);
static void gsf_pixmapsprite_draw         (GsfBaseSprite *s,
					   GsfPainter *p);
static GSList *gsf_pixmapsprite_approxat  (GsfBaseSprite *s);
static GsfCollisionMask *gsf_pixmapsprite_gcm(GsfBaseSprite  *s,
					      gint           *x,
					      gint           *y);

static GsfBaseSpriteClass *parent_class = NULL;

GtkType
gsf_pixmapsprite_get_type(void)
{
    static GtkType sprite_type = 0;

    if (!sprite_type) {
	GtkTypeInfo gsf_pixmapsprite_info = {
	    "GsfPixmapSprite",
	    sizeof(GsfPixmapSprite),
	    sizeof(GsfPixmapSpriteClass),
	    (GtkClassInitFunc) gsf_pixmapsprite_class_init,
	    (GtkObjectInitFunc) gsf_pixmapsprite_init,
	    (GtkArgSetFunc) NULL,
	    (GtkArgGetFunc) NULL,
	};

	sprite_type = gtk_type_unique(gsf_basesprite_get_type(),
				      &gsf_pixmapsprite_info);
    }

    return sprite_type;
}

static void
gsf_pixmapsprite_class_init(GsfPixmapSpriteClass *klass)
{
    GtkObjectClass *object_class;
    GsfBaseSpriteClass *basesprite_class;

    object_class = (GtkObjectClass*)klass;
    basesprite_class = (GsfBaseSpriteClass *)klass;
    parent_class = gtk_type_class(gsf_basesprite_get_type());

    object_class->destroy = gsf_pixmapsprite_destroy;
    basesprite_class->makevisible = gsf_pixmapsprite_makevisible;
    basesprite_class->draw = gsf_pixmapsprite_draw;
    basesprite_class->approxat = gsf_pixmapsprite_approxat;
    basesprite_class->getcollisionmask = gsf_pixmapsprite_gcm;
    basesprite_class->getnextcollisionmask = gsf_pixmapsprite_gcm;
}

static void
gsf_pixmapsprite_init(GsfPixmapSprite *sprite)
{
    sprite->x = 0;
    sprite->y = 0;
    sprite->currentframe = 0;
    sprite->numframes = 0;
    sprite->gfx = NULL;
}

GsfPixmapSprite *
gsf_pixmapsprite_new(void)
{
    return gtk_type_new(gsf_pixmapsprite_get_type());
}

void
gsf_pixmapsprite_ref(GsfPixmapSprite *pixmapsprite)
{
    g_return_if_fail(pixmapsprite!=NULL && GSF_IS_PIXMAPSPRITE(pixmapsprite));

    gtk_object_ref(GTK_OBJECT(pixmapsprite));
}

void
gsf_pixmapsprite_unref(GsfPixmapSprite *pixmapsprite)
{
    g_return_if_fail(pixmapsprite!=NULL && GSF_IS_PIXMAPSPRITE(pixmapsprite));

    gtk_object_unref(GTK_OBJECT(pixmapsprite));
}

static void
gsf_pixmapsprite_destroy(GtkObject *object)
{
    GsfPixmapSprite *pixmapsprite;

    g_return_if_fail(object != NULL && GSF_IS_PIXMAPSPRITE(object));

    pixmapsprite = GSF_PIXMAPSPRITE(object);
    gsf_pixmapsprite_set(pixmapsprite, 0, 0, 0, NULL, 0, NULL, 0);

    if (GTK_OBJECT_CLASS(parent_class)->destroy)
	(* GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}

void
gsf_pixmapsprite_set(GsfPixmapSprite *sprite,
		     gint x,
		     gint y,
		     gint frame,
		     GsfSpriteGfx **gfx,
		     gint numframes,
		     GsfSpriteField *f,
		     gint z)
{
    gint i;
    gboolean vis;
    GsfBaseSprite *basesprite;
    
    g_return_if_fail(sprite != NULL && GSF_IS_PIXMAPSPRITE(sprite));

    basesprite = GSF_BASESPRITE(sprite);
    
    vis = basesprite->visible;
    if (vis) gsf_basesprite_hide(basesprite);

    sprite->x = x;
    sprite->y = y;
    sprite->currentframe = frame < numframes ? frame : numframes-1;

    if (sprite->gfx) {
	for (i = 0; i < sprite->numframes; i++)
	    gsf_spritegfx_unref(sprite->gfx[i]);
	g_free(sprite->gfx);
	sprite->gfx = NULL;
	sprite->numframes = 0;
    }
    
    if (gfx && numframes) {
	sprite->numframes = numframes;
	sprite->gfx = g_new(GsfSpriteGfx *, numframes);
	for (i = 0; i < numframes; i++) {
	    sprite->gfx[i] = gfx[i];
	    gsf_spritegfx_ref(sprite->gfx[i]);
	    gtk_object_sink(GTK_OBJECT(sprite->gfx[i]));
	}
    }
    
    gsf_sprite_setz(sprite, z);
    gsf_sprite_setfield(sprite, f);
    
    if (vis || f) gsf_basesprite_show(basesprite);

    gsf_basesprite_dirtycollidecache(basesprite);
}

void
gsf_pixmapsprite_setxyf(GsfPixmapSprite *sprite,
			gint x,
			gint y,
			gint frame)
{
    GsfBaseSprite *basesprite;
    g_return_if_fail(sprite != NULL && GSF_IS_PIXMAPSPRITE(sprite));
    g_return_if_fail((guint)frame < (guint)sprite->numframes);
    
    if (x != sprite->x || y != sprite->y || frame != sprite->currentframe) {
	basesprite = GSF_BASESPRITE(sprite);

	if (basesprite->field && basesprite->visible) {
	    GsfRectangle oldarea, newarea;
	    GsfSpriteGfx *tmp1, *tmp2;
	    tmp1 = sprite->gfx[sprite->currentframe];
	    tmp2 = sprite->gfx[frame];

	    oldarea.x = sprite->x - tmp1->hotx;
	    oldarea.y = sprite->y - tmp1->hoty;
	    oldarea.width = tmp1->width;
	    oldarea.height = tmp1->height;

	    newarea.x = x - tmp2->hotx;
	    newarea.y = y - tmp2->hoty;
	    newarea.width = tmp2->width;
	    newarea.height = tmp2->height;
	    gsf_spritefield_movechanged(basesprite->field,
					&oldarea, &newarea,
					basesprite);
	}

	sprite->x = x;
	sprite->y = y;
	sprite->currentframe = frame;
	
	gsf_basesprite_dirtycollidecache(basesprite);
    }
}
		    
void
gsf_pixmapsprite_getsize(GsfPixmapSprite *sprite,
			 gint *width, gint *height)
{
    GsfSpriteGfx *g;
    g_return_if_fail(sprite != NULL && GSF_IS_PIXMAPSPRITE(sprite));
    g_return_if_fail(sprite->gfx);

    g = sprite->gfx[sprite->currentframe];
    *width = g->width;
    *height = g->height;
}

void
gsf_pixmapsprite_pastegfxrect(GsfPixmapSprite *dest,
			      GsfPixmapSprite *src,
			      gint destx,
			      gint desty,
			      GsfRectangle *srcarea,
			      gboolean maskedcopy)
{
    GsfBaseSprite *basesprite;
    gboolean vis;
    
    g_return_if_fail(dest && GSF_IS_PIXMAPSPRITE(dest));
    g_return_if_fail(src && GSF_IS_PIXMAPSPRITE(src));

    basesprite = GSF_BASESPRITE(dest);
    vis = basesprite->visible;
    if (vis) gsf_basesprite_hide(basesprite);
    
    gsf_spritegfx_pastegfxrect(dest->gfx[dest->currentframe],
			       src->gfx[src->currentframe],
			       destx, desty, srcarea, maskedcopy);

    if (vis) gsf_basesprite_show(basesprite);
}

void
gsf_pixmapsprite_pastegfx(GsfPixmapSprite *dest,
			  GsfPixmapSprite *src,
			  gint destx,
			  gint desty)
{
    GsfBaseSprite *basesprite;
    gboolean vis;
    
    g_return_if_fail(dest && GSF_IS_PIXMAPSPRITE(dest));
    g_return_if_fail(src && GSF_IS_PIXMAPSPRITE(src));

    basesprite = GSF_BASESPRITE(dest);
    vis = basesprite->visible;
    if (vis) gsf_basesprite_hide(basesprite);
    
    gsf_spritegfx_pastegfx(dest->gfx[dest->currentframe],
			   src->gfx[src->currentframe],
			   destx, desty);

    if (vis) gsf_basesprite_show(basesprite);
}

void
gsf_pixmapsprite_clearrect(GsfPixmapSprite *dest,
			   GsfRectangle *area)
{
    GsfBaseSprite *basesprite;
    gboolean vis;
    
    g_return_if_fail(dest && GSF_IS_PIXMAPSPRITE(dest));

    basesprite = GSF_BASESPRITE(dest);
    vis = basesprite->visible;
    if (vis) gsf_basesprite_hide(basesprite);
    
    gsf_spritegfx_clearrect(dest->gfx[dest->currentframe], area);

    if (vis) gsf_basesprite_show(basesprite);
}



static void
gsf_pixmapsprite_makevisible(GsfBaseSprite *s,
			     gboolean vis)
{
    GsfSpriteGfx *tmp;
    GsfPixmapSprite *gps;
    
    g_return_if_fail(s != NULL && GSF_IS_PIXMAPSPRITE(s));
    
    if (s->field) {
	GsfRectangle area;
	
	gps = GSF_PIXMAPSPRITE(s);
	tmp = gps->gfx[gps->currentframe];

	area.x = gps->x - tmp->hotx;
	area.y = gps->y - tmp->hoty;
	area.width = tmp->width;
	area.height = tmp->height;
	
	gsf_spritefield_setchanged(s->field, &area, s, vis);
    }
}

static void
gsf_pixmapsprite_draw(GsfBaseSprite *s,
		      GsfPainter *p)
{
    GsfRectangle area;
    GsfSpriteGfx *tmp;
    GsfPixmapSprite *gps;
    g_return_if_fail(s != NULL && GSF_IS_PIXMAPSPRITE(s));

    gps = GSF_PIXMAPSPRITE(s);
    tmp = gps->gfx[gps->currentframe];

    area.x = gps->x - tmp->hotx;
    area.y = gps->y - tmp->hoty;
    area.width = (guint)-1;
    area.height = (guint)-1;
    gsf_painter_draw_masked_pixmap(p, tmp->pixmap, tmp->mask, &area);
/*     g_message("draw: (%d, %d, %u, %u)/(%d %d %d)", */
/* 	      ts->x, ts->y, ts->w, ts->h, */
/* 	      ts->c.red, ts->c.green, ts->c.blue); */
}

static GSList *
gsf_pixmapsprite_approxat(GsfBaseSprite *s)
{
    GsfSpriteGfx *tmp;
    GsfPixmapSprite *gps;
    GsfRectangle area;
    GsfSpriteField *f;
    
    g_return_val_if_fail(s != NULL && GSF_IS_PIXMAPSPRITE(s), NULL);
    if (!s->field)
	return NULL;

    gps = GSF_PIXMAPSPRITE(s);    
    tmp = gps->gfx[gps->currentframe];

    area.x = gps->x - tmp->hotx;
    area.y = gps->y - tmp->hoty;
    area.width = tmp->width;
    area.height = tmp->height;
    
    return gsf_spritefield_approx_at_rect(s->field, &area);
}

static GsfCollisionMask *
gsf_pixmapsprite_gcm(GsfBaseSprite *s,
		     gint          *x,
		     gint          *y)
{
    GsfPixmapSprite *ps;
    GsfSpriteGfx *tmp;
    g_return_val_if_fail(s != NULL && GSF_IS_PIXMAPSPRITE(s), NULL);

    ps = GSF_PIXMAPSPRITE(s);
    if (!ps->gfx)
	return NULL;
    
    tmp = ps->gfx[ps->currentframe];
    if (x)
	*x = ps->x - tmp->hotx;
    if (y)
	*y = ps->y - tmp->hoty;
    return tmp->collmask;
}

