/* GSF - GTK+ SpriteField.  Library to use sprites in GTK+
 * Copyright (C) 1998  Tore Skjellnes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include "gsfspritefield.h"
#include "gsfbasesprite.h"
#include "gsfspriteview.h"
#include "gsfclusterizer.h"

#define DEFAULT_CELLSIZE 20

static void gsf_spritefield_class_init (GsfSpriteFieldClass *klass);
static void gsf_spritefield_init       (GsfSpriteField      *spritefield);
static void gsf_spritefield_destroy    (GtkObject           *object);

static void gsf_spritefield_drawgraphics(GsfSpriteField *field,
					 GsfRectangle *area,
					 GSList *sprites,
					 GSList *views);
static void gsf_spritefield_bg_color    (GsfSpriteField *field,
					 GsfPainter *p,
					 GsfRectangle *area);
static void gsf_spritefield_bg_pixmap   (GsfSpriteField *field,
					 GsfPainter *p,
					 GsfRectangle *area);
static void gsf_spritefield_fg_ignore   (GsfSpriteField *field,
					 GsfPainter *p,
					 GsfRectangle *area);

static GtkObjectClass *parent_class = NULL;


GtkType
gsf_spritefield_get_type(void)
{
    static GtkType spritefield_type = 0;

    if (!spritefield_type) {
	GtkTypeInfo spritefield_info = {
	    "GsfSpriteField",
	    sizeof(GsfSpriteField),
	    sizeof(GsfSpriteFieldClass),
	    (GtkClassInitFunc) gsf_spritefield_class_init,
	    (GtkObjectInitFunc) gsf_spritefield_init,
	    (GtkArgSetFunc) NULL,
	    (GtkArgGetFunc) NULL,
	};

	spritefield_type = gtk_type_unique(gtk_object_get_type(),
					   &spritefield_info);
    }

    return spritefield_type;
}

static void
gsf_spritefield_class_init(GsfSpriteFieldClass *klass)
{
    GtkObjectClass *object_class;

    object_class = (GtkObjectClass*)klass;
    parent_class = gtk_type_class(gtk_object_get_type());

    object_class->destroy = gsf_spritefield_destroy;
    klass->drawgraphics = gsf_spritefield_drawgraphics;
}

static void
gsf_spritefield_init(GsfSpriteField *spritefield)
{
    spritefield->width = 0;
    spritefield->height = 0;
    spritefield->cellcols = 0;
    spritefield->cellrows = 0;
    spritefield->cellsize = 0;
    spritefield->cells = NULL;
    spritefield->all_objects = NULL;
    spritefield->view_list = NULL;
    spritefield->bg_color.pixel = 0;
    spritefield->bg_color.red = 0;
    spritefield->bg_color.green = 0;
    spritefield->bg_color.blue = 0;
    spritefield->bg_pixmap = NULL;
    spritefield->drawbackground = gsf_spritefield_bg_color;
    spritefield->drawforeground = gsf_spritefield_fg_ignore;
}

GsfSpriteField *
gsf_spritefield_new(guint width,
		    guint height,
		    guint cellsize)
{
    GsfSpriteField *spritefield;

    spritefield = gtk_type_new(gsf_spritefield_get_type());
    gsf_spritefield_resize_c(spritefield, width, height, cellsize);

    return spritefield;
}

void
gsf_spritefield_ref(GsfSpriteField *f)
{
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    gtk_object_ref(GTK_OBJECT(f));
}

void
gsf_spritefield_unref(GsfSpriteField *f)
{
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    gtk_object_unref(GTK_OBJECT(f));
}

static void
gsf_spritefield_destroy(GtkObject *object)
{
    GsfSpriteField *spritefield;

    g_return_if_fail(object != NULL && GSF_IS_SPRITEFIELD(object));

    spritefield = GSF_SPRITEFIELD(object);

    gsf_spritefield_removeallsprites(spritefield);
    gsf_spritefield_removeallviews(spritefield);

    if (spritefield->cells)
	g_free(spritefield->cells);
    spritefield->cells = NULL;

    if (spritefield->bg_pixmap) {
	gdk_pixmap_unref(spritefield->bg_pixmap);
	spritefield->bg_pixmap = NULL;
    }

    if (GTK_OBJECT_CLASS(parent_class)->destroy)
	(* GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}

void
gsf_spritefield_removeallsprites(GsfSpriteField *spritefield)
{
    GSList *iter;
    GSList *next;
    
    g_return_if_fail(spritefield != NULL && GSF_IS_SPRITEFIELD(spritefield));

    iter = spritefield->all_objects;
    while (iter) {
	next = iter->next;
	gsf_sprite_setfield(iter->data, NULL);
	iter = next;
    }
    
     g_slist_free(spritefield->all_objects);
     spritefield->all_objects = NULL;

/*     cells = spritefield->cells; */
/*     endcells = cells + spritefield->cellcols*spritefield->cellrows; */
/*     while (cells < endcells) { */
/* 	g_slist_free(cells->sprites); */
/* 	cells->sprites = NULL; */
/* 	cells->clean = FALSE; */
	
/* 	cells++; */
/*     } */
}

void
gsf_spritefield_removeallviews(GsfSpriteField *spritefield)
{
    GSList *iter;
    GSList *next;
    
    g_return_if_fail(spritefield != NULL && GSF_IS_SPRITEFIELD(spritefield));
    
    iter = spritefield->view_list;
    while (iter) {
	next = iter->next;
	gsf_spriteview_setfield(GSF_SPRITEVIEW(iter->data), NULL);
	iter = next;
    }
    g_slist_free(spritefield->view_list);
    spritefield->view_list = NULL;
}

void
gsf_spritefield_resize_c(GsfSpriteField *field,
			 guint width, guint height, guint cellsize)
{
    GSList *vis_objects;
    GSList *iter;
    
    g_return_if_fail(field != NULL && GSF_IS_SPRITEFIELD(field));
    g_return_if_fail(width > 0 && height > 0 && cellsize > 0);

    vis_objects = NULL;
    iter = field->all_objects;
    while (iter) {
	GsfBaseSprite *s = GSF_BASESPRITE(iter->data);
	if (s->visible) {
	    gsf_basesprite_hide(s);

	    vis_objects = g_slist_prepend(vis_objects, s);
	}
	iter = iter->next;
    }

    field->width = width;
    field->height = height;
    field->cellsize = cellsize;
    field->cellcols = (width + cellsize - 1)/cellsize;
    field->cellrows = (height + cellsize - 1)/cellsize;
    field->cellsize = cellsize;
    if (field->cells)
	g_free(field->cells);
    field->cells = g_new0(GsfSpriteCell, field->cellcols*field->cellrows);

    iter = vis_objects;
    while (iter) {
	GsfBaseSprite *s = (GsfBaseSprite *)iter->data;
	gsf_basesprite_show(s);
	
	iter = iter->next;
    }
    g_slist_free(vis_objects);
}

void
gsf_spritefield_resize(GsfSpriteField *field, guint width, guint height)
{
    guint cellsize;
    g_return_if_fail(field != NULL && GSF_IS_SPRITEFIELD(field));

    cellsize = field->cellsize ? field->cellsize : DEFAULT_CELLSIZE;

    gsf_spritefield_resize_c(field, width, height, cellsize);
}


void
gsf_spritefield_setbgcolor(GsfSpriteField *f, GdkColor *c)
{
    f->bg_color = *c;
    f->drawbackground = gsf_spritefield_bg_color;
}

void
gsf_spritefield_setbgpixmap(GsfSpriteField *f, GdkPixmap *p)
{
    if (f->bg_pixmap)
	gdk_pixmap_unref(f->bg_pixmap);
    f->bg_pixmap = gdk_pixmap_ref(p);
    f->drawbackground = gsf_spritefield_bg_pixmap;
}

void
gsf_spritefield_setbgfunction(GsfSpriteField *f, GsfSpriteFieldDrawFunc d)
{
    f->drawbackground = d;
}

void
gsf_spritefield_setfgfunction(GsfSpriteField *f, GsfSpriteFieldDrawFunc d)
{
    f->drawforeground = d;
}



static gboolean
gsf_slist_remove(GSList **list, gpointer data)
{
    GSList *prev = NULL;
    GSList *iter = *list;
    while (iter && iter->data != data) {
	prev = iter;
	iter = iter->next;
    }
    if (!iter)
	return FALSE;

    if (prev)
	prev->next = iter->next;
    else
	*list = iter->next;

    g_slist_free_1(iter);
    return TRUE;
}

void
gsf_spritefield_addsprite(GsfSpriteField *f, GsfBaseSprite *s)
{
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    g_return_if_fail(s != NULL && GSF_IS_BASESPRITE(s));

    if (s->field) {
	g_warning("Adding sprite belonging to another spritefield");
	gsf_spritefield_removesprite(s->field, s);
    }
    
    s->field = f;
    f->all_objects = g_slist_prepend(f->all_objects, s);
    gsf_basesprite_ref(s);
    gtk_object_sink(GTK_OBJECT(s));
/*     gtk_object_ref(GTK_OBJECT(f)); */
}

void
gsf_spritefield_removesprite(GsfSpriteField *f, GsfBaseSprite *s)
{
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    g_return_if_fail(s != NULL && GSF_IS_BASESPRITE(s));

    if (s->field == f) {
	if (gsf_sprite_isvisible(s))
	    gsf_basesprite_hide(s);
	if (gsf_slist_remove(&f->all_objects, s)) {       
	    s->field = NULL;
	    gsf_basesprite_unref(s);
	}
/* 	gtk_object_unref(GTK_OBJECT(f)); */
    } else {
	g_warning("Trying to remove alien sprite from list");
	return;
    }
}

void
gsf_spritefield_addview(GsfSpriteField *f, GsfSpriteView *v)
{
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    g_return_if_fail(v != NULL && GSF_IS_SPRITEVIEW(v));

    if (v->field) {
	g_warning("Adding view belonging to another spritefield");
	gsf_spritefield_removeview(v->field, v);
    }
    
    v->field = f;
    f->view_list = g_slist_prepend(f->view_list, v);
    gsf_spritefield_ref(f);
    gtk_object_sink(GTK_OBJECT(f));
    
    gtk_widget_ref(GTK_WIDGET(v));
    gtk_object_sink(GTK_OBJECT(v));
}

void
gsf_spritefield_removeview(GsfSpriteField *f, GsfSpriteView *v)
{
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    g_return_if_fail(v != NULL && GSF_IS_SPRITEVIEW(v));

    if (v->field == f && gsf_slist_remove(&f->view_list, v)) {
	v->field = NULL;
	gtk_widget_unref(GTK_WIDGET(v));
	gsf_spritefield_unref(f);
    } else {
	g_warning("Trying to remove alien view from list");
	return;
    }
}

static gboolean
gsf_spritefield_world_to_cell_area(GsfSpriteField *f,
				   GsfRectangle *area,
				   gint *left,
				   gint *top,
				   gint *right,
				   gint *bottom)
{
   /* Note: (-5)/6 = 0! */
    *right = area->x + area->width - 1;
    if (*right < 0)
	return FALSE;
    *right /= f->cellsize;
    if (*right >= f->cellcols)
	*right = f->cellcols - 1;
    
    *left = area->x;
    if (*left < 0)
	*left = 0;
    *left /= f->cellsize;
    if (*left >= f->cellcols)
	return FALSE;
    
    *bottom = area->y + area->height - 1;
    if (*bottom < 0)
	return FALSE;
    *bottom /= f->cellsize;
    if (*bottom >= f->cellrows)
	*bottom = f->cellrows - 1;
    
    *top = area->y;
    if (*top < 0)
	*top = 0;
    *top /= f->cellsize;
    if (*top >= f->cellrows)
	return FALSE;

    return TRUE;
}

void
gsf_spritefield_setchanged(GsfSpriteField *f,
			   GsfRectangle *area,
			   GsfBaseSprite *s,
			   gboolean add)
{
    gint left, top, right, bottom, i, j;
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));

    if (!gsf_spritefield_world_to_cell_area(f, area,
					    &left, &top, &right, &bottom))
	return;
    
    if (s)
	if (add)
	    for (j = top; j <= bottom; j++)
		for (i = left; i <= right; i++) {
		    GsfSpriteCell *c = &f->cells[j*f->cellcols + i];
		    c->sprites = g_slist_prepend(c->sprites, s);
		    c->clean = FALSE;
		}
	else
	    for (j = top; j <= bottom; j++)
		for (i = left; i <= right; i++) {
		    GsfSpriteCell *c = &f->cells[j*f->cellcols + i];
		    c->sprites = g_slist_remove(c->sprites, s);
		    c->clean = FALSE;
		}
    else
	for (j = top; j <= bottom; j++)
	    for (i = left; i <= right; i++)
		f->cells[j*f->cellcols + i].clean = FALSE;
}


void
gsf_spritefield_movechanged(GsfSpriteField *f,
			    GsfRectangle *oldarea,
			    GsfRectangle *newarea,
			    GsfBaseSprite *s)
{
    gint oldleft, oldtop, oldright, oldbottom;
    gint newleft, newtop, newright, newbottom;
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    g_return_if_fail(s != NULL && GSF_IS_BASESPRITE(s));

    if (!gsf_spritefield_world_to_cell_area(f, oldarea,
					    &oldleft, &oldtop,
					    &oldright, &oldbottom)) {
	gsf_spritefield_setchanged(f, newarea, s, TRUE);
	return;
    }
    if (!gsf_spritefield_world_to_cell_area(f, newarea,
					    &newleft, &newtop,
					    &newright, &newbottom)) {
	gsf_spritefield_setchanged(f, oldarea, s, FALSE);
	return;
    }
    
    /* Add new top */
    while (newtop < oldtop && newtop <= newbottom) {
	GsfSpriteCell *c = &f->cells[newtop*f->cellcols + newleft];
	GsfSpriteCell *ec = &f->cells[newtop*f->cellcols + newright];
	while (c <= ec) {
	    c->sprites = g_slist_prepend(c->sprites, s);
	    c->clean = FALSE;
	    c++;
	}
	newtop++;
    }

    /* Remove old top */
    while (oldtop < newtop && oldtop <= oldbottom) {
	GsfSpriteCell *c = &f->cells[oldtop*f->cellcols + oldleft];
	GsfSpriteCell *ec = &f->cells[oldtop*f->cellcols + oldright];
	while (c <= ec) {
	    c->sprites = g_slist_remove(c->sprites, s);
	    c->clean = FALSE;
	    c++;
	}
	oldtop++;
    }

    /* Add new bottom */
    while (newbottom > oldbottom && newbottom >= newtop) {
	GsfSpriteCell *c = &f->cells[newbottom*f->cellcols + newleft];
	GsfSpriteCell *ec = &f->cells[newbottom*f->cellcols + newright];
	while (c <= ec) {
	    c->sprites = g_slist_prepend(c->sprites, s);
	    c->clean = FALSE;
	    c++;
	}
	newbottom--;
    }

    /* Remove old bottom */
    while (oldbottom > newbottom && oldbottom >= oldtop) {
	GsfSpriteCell *c = &f->cells[oldbottom*f->cellcols + oldleft];
	GsfSpriteCell *ec = &f->cells[oldbottom*f->cellcols + oldright];
	while (c <= ec) {
	    c->sprites = g_slist_remove(c->sprites, s);
	    c->clean = FALSE;
	    c++;
	}
	oldbottom--;
    }

    /* Add new left side */
    while (newleft < oldleft && newleft <= newright) {
	GsfSpriteCell *c = &f->cells[newtop*f->cellcols + newleft];
	GsfSpriteCell *ec = &f->cells[newbottom*f->cellcols + newleft];
	while (c <= ec) {
	    c->sprites = g_slist_prepend(c->sprites, s);
	    c->clean = FALSE;
	    c += f->cellcols;
	}
	newleft++;
    }

    /* Remove old left side */
    while (oldleft < newleft && oldleft <= oldright) {
	GsfSpriteCell *c = &f->cells[oldtop*f->cellcols + oldleft];
	GsfSpriteCell *ec = &f->cells[oldbottom*f->cellcols + oldleft];
	while (c <= ec) {
	    c->sprites = g_slist_remove(c->sprites, s);
	    c->clean = FALSE;
	    c += f->cellcols;
	}
	oldleft++;
    }

    /* Add new right side */
    while (newright > oldright && newright >= newleft) {
	GsfSpriteCell *c = &f->cells[newtop*f->cellcols + newright];
	GsfSpriteCell *ec = &f->cells[newbottom*f->cellcols + newright];
	while (c <= ec) {
	    c->sprites = g_slist_prepend(c->sprites, s);
	    c->clean = FALSE;
	    c += f->cellcols;
	}
	newright--;
    }

    /* Remove old right side */
    while (oldright > newright && oldright >= oldleft) {
	GsfSpriteCell *c = &f->cells[oldtop*f->cellcols + oldright];
	GsfSpriteCell *ec = &f->cells[oldbottom*f->cellcols + oldright];
	while (c <= ec) {
	    c->sprites = g_slist_remove(c->sprites, s);
	    c->clean = FALSE;
	    c += f->cellcols;
	}
	oldright--;
    }

    /* Clean the clean flag in the overlapping region */
    while (newtop <= newbottom) {
	GsfSpriteCell *c = &f->cells[newtop*f->cellcols + newleft];
	GsfSpriteCell *ec = &f->cells[newtop*f->cellcols + newright];
	while (c <= ec) {
	    c->clean = FALSE;
	    c++;
	}
	newtop++;
    }
}


void
gsf_spritefield_update(GsfSpriteField *f)
{
    GsfClusterizer *clusterizer;
    GSList *iter;
    GsfRectangle *area;

    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    
    clusterizer = gsf_clusterizer_new();
    iter = f->view_list;
    while (iter) {
	GsfSpriteViewClass *v;
	v = GSF_SPRITEVIEW_CLASS(GTK_OBJECT(iter->data)->klass);

	area = GSF_RECTANGLE_NEW();
	v->getviewarea(GSF_SPRITEVIEW(iter->data), area);
	if (area->width > 0 && area->height > 0)
	    gsf_clusterizer_add(clusterizer, area);
	iter = iter->next;
    }
    
    while ( (area = gsf_clusterizer_pop(clusterizer)) ) {
	gsf_spritefield_drawareaviews(f, FALSE, area, f->view_list);
	GSF_RECTANGLE_FREE(area);
    }

    gsf_clusterizer_unref(clusterizer);
}

void
gsf_spritefield_updateview(GsfSpriteField *f,
			   GsfRectangle *area,
			   GsfSpriteView *v)
{
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    g_return_if_fail(v != NULL && GSF_IS_SPRITEVIEW(v));

    gsf_spritefield_drawareaview(f, TRUE, area, v);
}





static gint
gsf_sprite_z_compare(gconstpointer a, gconstpointer b)
{
    gint r = ((GsfBaseSprite *)a)->z - ((GsfBaseSprite *)b)->z;
    return r ? r : GPOINTER_TO_INT(a) - GPOINTER_TO_INT(b);
}

static GSList *
gsf_spritefield_at_rect_unchecked(GsfSpriteField *f,
				  GsfRectangle *area)
{
    /* Note: reference count not updated for sprites.  nbd. */
    GSList *sprites, *iter;
    gint mincol, minrow, maxcol, maxrow, row, col;
    
    sprites = NULL;
    mincol = area->x/f->cellsize;
    minrow = area->y/f->cellsize;
    maxcol = (area->x + area->width - 1)/f->cellsize;
    maxrow = (area->y + area->height - 1)/f->cellsize;
    for (row = minrow; row <= maxrow; row++)
	for (col = mincol; col <= maxcol; col++)
	    for (iter = f->cells[row*f->cellcols + col].sprites;
		 iter;
		 iter = iter->next)
		if (!((GsfBaseSprite *)(iter->data))->tag) {
		    ((GsfBaseSprite *)(iter->data))->tag = TRUE;
		    sprites = g_slist_insert_sorted(sprites,
						    iter->data,
						    gsf_sprite_z_compare);
		}
    for (iter = sprites; iter; iter = iter->next)
	((GsfBaseSprite *)(iter->data))->tag = FALSE;
    
    return sprites;
}



void
gsf_spritefield_drawareaviews(GsfSpriteField *f,
			      gboolean force,
			      GsfRectangle *area,
			      GSList *views)
{
    GsfRectangle realarea;
    gint mincol, minrow, maxcol, maxrow, col, row;

    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));

    realarea.x = 0;
    realarea.y = 0;
    realarea.width = f->width;
    realarea.height = f->height;

    /* clip area to spritefield area */
    gsf_rectangle_intersection(&realarea, &realarea, area);
    if (gsf_rectangle_isempty(&realarea))
	return;
    
    if (!force) {
	GsfRectangle *clusterarea;
	GsfClusterizer *clusterizer;
	
	clusterizer = gsf_clusterizer_new();
	mincol = realarea.x/f->cellsize;
	minrow = realarea.y/f->cellsize;
	maxcol = (realarea.x + realarea.width - 1)/f->cellsize;
	maxrow = (realarea.y + realarea.height - 1)/f->cellsize;
	for (row = minrow; row <= maxrow; row++)
	    for (col = mincol; col <= maxcol; col++)
		if (!f->cells[row*f->cellcols + col].clean) {
		    f->cells[row*f->cellcols + col].clean = TRUE;
		    gsf_clusterizer_add_point(clusterizer, col, row);
		}
	
	while ( (clusterarea = gsf_clusterizer_pop(clusterizer)) ) {
	    clusterarea->x *= f->cellsize;
	    clusterarea->y *= f->cellsize;
	    clusterarea->width *= f->cellsize;
	    clusterarea->height *= f->cellsize;
	    gsf_spritefield_drawareaviews(f, TRUE, clusterarea, views);
	    GSF_RECTANGLE_FREE(clusterarea);
	}
	
	gsf_clusterizer_unref(clusterizer);
    } else {
	GsfSpriteFieldClass *sfc;
	GSList *sprites;
	
	if (views == f->view_list) {
	    mincol = (realarea.x + f->cellsize - 1)/f->cellsize;
	    minrow = (realarea.y + f->cellsize - 1)/f->cellsize;
	    maxcol = (realarea.x + realarea.width)/f->cellsize - 1;
	    maxrow = (realarea.y + realarea.height)/f->cellsize - 1;
	    for (row = minrow; row <= maxrow; row++)
		for (col = mincol; col <= maxcol; col++)
		    f->cells[row*f->cellcols + col].clean = TRUE;
	}
	sprites = gsf_spritefield_at_rect_unchecked(f, &realarea);

	sfc = GSF_SPRITEFIELD_CLASS(GTK_OBJECT(f)->klass);
	sfc->drawgraphics(f, &realarea, sprites, views);
	g_slist_free(sprites);
    }
}


void
gsf_spritefield_drawarea(GsfSpriteField *f,
			 gboolean force,
			 GsfRectangle *area)
{
    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    gsf_spritefield_drawareaviews(f, force, area, f->view_list);
}

void
gsf_spritefield_drawareaview(GsfSpriteField *f,
			     gboolean force,
			     GsfRectangle *area,
			     GsfSpriteView *v)
{
    GSList *l;

    g_return_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f));
    g_return_if_fail(v != NULL && GSF_IS_SPRITEVIEW(v));

    l = g_slist_prepend(NULL, v);
    gsf_spritefield_drawareaviews(f, force, area, l);
    g_slist_free(l);
}


static void
gsf_spritefield_drawgraphics(GsfSpriteField *field,
			     GsfRectangle *area,
			     GSList *sprites,
			     GSList *views)
{
    GsfSpriteFieldClass *fc;
    fc = GSF_SPRITEFIELD_CLASS(GTK_OBJECT(field)->klass);
    while (views) {
	GsfPainter *p;
	GSList *iter;

	p = gsf_painter_new();
	gsf_painter_init_pixmap(p, GTK_WIDGET(views->data),
				area->width, area->height);
	gsf_painter_set_origin(p, -area->x, -area->y);
	
	(* field->drawbackground)(field, p, area);
	iter = sprites;
	while (iter) {
	    GsfBaseSprite *sprite;
	    GsfBaseSpriteClass *sc;

	    sprite = GSF_BASESPRITE(iter->data);
	    sc = GSF_BASESPRITE_CLASS(GTK_OBJECT(sprite)->klass);

	    (* sc->draw)(sprite, p);
	    
	    iter = iter->next;
	}
	(* field->drawforeground)(field, p, area);

	gdk_draw_pixmap(GTK_WIDGET(views->data)->window,
			p->gc,
			p->drawable,
			0, 0,
			area->x, area->y,
			area->width, area->height);

	gsf_painter_unref(p);
	
	views = views->next;
    }
}

static void
gsf_spritefield_bg_color(GsfSpriteField *field,
			 GsfPainter *p,
			 GsfRectangle *area)
{
    gsf_painter_set_foreground(p, &field->bg_color);
    gsf_painter_fill_rectangle(p, area);
}

static void
gsf_spritefield_bg_pixmap(GsfSpriteField *field,
			  GsfPainter *p,
			  GsfRectangle *area)
{
    gsf_painter_draw_tiled_pixmap(p, field->bg_pixmap, 0, 0, area);
}

static void
gsf_spritefield_fg_ignore(GsfSpriteField *field,
			  GsfPainter *p,
			  GsfRectangle *area)
{
}




GSList *
gsf_spritefield_approx_at(GsfSpriteField *f,
			  gint x,
			  gint y)
{
    GsfRectangle area;
    g_return_val_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f), NULL);

    if ((guint)x >= (guint)f->width || (guint)y >= (guint)f->height)
	return NULL;

    area.x = x;
    area.y = y;
    area.width = 1;
    area.height = 1;
    return gsf_spritefield_at_rect_unchecked(f, &area);
}

GSList *
gsf_spritefield_approx_at_rect(GsfSpriteField *f,
			       GsfRectangle *area)
{
    GsfRectangle realarea;
    g_return_val_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f), NULL);

    realarea.x = 0;
    realarea.y = 0;
    realarea.width = f->width;
    realarea.height = f->height;

    /* clip area to spritefield area */
    gsf_rectangle_intersection(&realarea, &realarea, area);
    if (gsf_rectangle_isempty(&realarea))
	return NULL;

    return gsf_spritefield_at_rect_unchecked(f, &realarea);
}
			       
GSList *
gsf_spritefield_at(GsfSpriteField *f,
		   gint x,
		   gint y)
{
    GSList *prev;
    GSList *iter;
    GSList *list;

    prev = NULL;
    list = iter = gsf_spritefield_approx_at(f, x, y);
    while (iter) {
	GSList *next = iter->next;
	if (!gsf_sprite_spriteat(iter->data, x, y)) {
	    if (prev)
		prev->next = next;
	    else
		list = next;
	    g_slist_free_1(iter);
	} else
	    prev = iter;
	iter = next;
    }
    
    return list;
}

GSList *
gsf_spritefield_get_collisions(GsfSpriteField *f)
{
    GSList *list;
    GSList *iter;
    g_return_val_if_fail(f != NULL && GSF_IS_SPRITEFIELD(f), NULL);

    list = NULL;
    iter = f->all_objects;
    while (iter) {
	GsfBaseSprite *s = GSF_BASESPRITE(iter->data);	
	if (gsf_basesprite_at(s))
	    list = g_slist_prepend(list, s);
	    
	iter = iter->next;
    }

    return list;
}
