/* GSF - GTK+ SpriteField.  Library to use sprites in GTK+
 * Copyright (C) 1998  Tore Skjellnes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include "gsfspritegfx.h"

static void gsf_spritegfx_class_init   (GsfSpriteGfxClass *klass);
static void gsf_spritegfx_init         (GsfSpriteGfx      *pixmap);
static void gsf_spritegfx_destroy      (GtkObject         *object);
static void gsf_spritegfx_p2c_nop      (gint              *x,
					gint              *y);
static void gsf_spritegfx_p2c_div      (gint              *x,
					gint              *y);
static void gsf_spritegfx_p2c_mul      (gint              *x,
					gint              *y);
static void gsf_spritegfx_p2c_muldiv   (gint              *x,
					gint              *y);

static GtkObjectClass *parent_class = NULL;

GtkType
gsf_spritegfx_get_type(void)
{
    static GtkType spritegfx_type = 0;

    if (!spritegfx_type) {
	GtkTypeInfo spritegfx_info = {
	    "GsfSpriteGfx",
	    sizeof(GsfSpriteGfx),
	    sizeof(GsfSpriteGfxClass),
	    (GtkClassInitFunc) gsf_spritegfx_class_init,
	    (GtkObjectInitFunc) gsf_spritegfx_init,
	    (GtkArgSetFunc) NULL,
	    (GtkArgGetFunc) NULL,
	};

	spritegfx_type = gtk_type_unique(gtk_object_get_type(),
					 &spritegfx_info);
    }

    return spritegfx_type;
}

static void
gsf_spritegfx_class_init(GsfSpriteGfxClass *klass)
{
    GtkObjectClass *object_class;

    object_class = (GtkObjectClass*)klass;
    parent_class = gtk_type_class(gtk_object_get_type());

    object_class->destroy = gsf_spritegfx_destroy;
    klass->pixmap_to_coll = gsf_spritegfx_p2c_nop;
}

static void
gsf_spritegfx_init(GsfSpriteGfx *spritegfx)
{
    spritegfx->pixmap = NULL;
    spritegfx->width = 0;
    spritegfx->height = 0;
    spritegfx->hotx = 0;
    spritegfx->hoty = 0;
    spritegfx->collmask = NULL;
}

GsfSpriteGfx *
gsf_spritegfx_new(void)
{
    return gtk_type_new(gsf_spritegfx_get_type());
}

void
gsf_spritegfx_ref(GsfSpriteGfx *spritegfx)
{
    g_return_if_fail(spritegfx != NULL && GSF_IS_SPRITEGFX(spritegfx));

    gtk_object_ref(GTK_OBJECT(spritegfx));
}

void
gsf_spritegfx_unref(GsfSpriteGfx *spritegfx)
{
    g_return_if_fail(spritegfx != NULL && GSF_IS_SPRITEGFX(spritegfx));

    gtk_object_unref(GTK_OBJECT(spritegfx));
}

static void
gsf_spritegfx_destroy(GtkObject *object)
{
    GsfSpriteGfx *spritegfx;
    g_return_if_fail(object != NULL && GSF_IS_SPRITEGFX(object));

    spritegfx = GSF_SPRITEGFX(object);

    if (spritegfx->pixmap) {
	gdk_pixmap_unref(spritegfx->pixmap);
	spritegfx->pixmap = NULL;
    }
    if (spritegfx->mask) {
	gdk_pixmap_unref(spritegfx->mask);
	spritegfx->mask = NULL;
    }
    if (spritegfx->collmask) {
	gsf_collisionmask_unref(spritegfx->collmask);
	spritegfx->collmask = NULL;
    }
    
    if (GTK_OBJECT_CLASS(parent_class)->destroy)
	(* GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}


static void
gsf_spritegfx_set_priv(GsfSpriteGfx *spritegfx,
		       GdkPixmap *pixmap,
		       GdkBitmap *mask)
{
    gint width, height;
    gint maskwidth, maskheight;
    g_return_if_fail(spritegfx != NULL && GSF_IS_SPRITEGFX(spritegfx));

    gdk_window_get_size(pixmap, &width, &height);
    gdk_window_get_size(mask, &maskwidth, &maskheight);
    if (width != maskwidth || height != maskheight)
	g_warning("pixmap and mask has different sizes in gsf_spritegfx: "
		  "(%d x %d) != (%d x %d)",
		  width, height, maskwidth, maskheight);

    spritegfx->width = width;
    spritegfx->height = height;
    
    if (spritegfx->pixmap)
	gdk_pixmap_unref(spritegfx->pixmap);
    spritegfx->pixmap = gdk_pixmap_ref(pixmap);

    if (spritegfx->mask)
	gdk_pixmap_unref(spritegfx->mask);
    spritegfx->mask = gdk_pixmap_ref(mask);

    if (spritegfx->collmask)
	gsf_collisionmask_unref(spritegfx->collmask);
    spritegfx->collmask = gsf_collisionmask_new();
    gsf_collisionmask_ref(spritegfx->collmask);
    gtk_object_sink(GTK_OBJECT(spritegfx->collmask));
}

void
gsf_spritegfx_set_coll(GsfSpriteGfx *spritegfx,
		       GdkPixmap *pixmap,
		       GdkBitmap *mask,
		       GdkImage *collmask)
{
    gsf_spritegfx_set_priv(spritegfx, pixmap, mask);
    gsf_collisionmask_set_coll(spritegfx->collmask, collmask);
}

void
gsf_spritegfx_set(GsfSpriteGfx *spritegfx,
		  GdkPixmap *pixmap,
		  GdkBitmap *mask)
{
    gsf_spritegfx_set_priv(spritegfx, pixmap, mask);
    gsf_collisionmask_set(spritegfx->collmask, mask);
    gsf_spritegfx_scale_collmask(spritegfx);
}

gboolean
gsf_spritegfx_collides(GsfSpriteGfx *spritegfx1,
		       gint x1,
		       gint y1,
		       GsfSpriteGfx *spritegfx2,
		       gint x2,
		       gint y2)
{
    x1 -= spritegfx1->hotx;
    y1 -= spritegfx1->hoty;
    gsf_spritegfx_p2c(spritegfx1, &x1, &y1);
    x2 -= spritegfx2->hotx;
    y2 -= spritegfx2->hoty;
    gsf_spritegfx_p2c(spritegfx2, &x2, &y2);

    return gsf_collisionmask_collides(spritegfx1->collmask, x1, y1,
				      spritegfx2->collmask, x2, y2);
}

gboolean
gsf_spritegfx_collides_at(GsfSpriteGfx *spritegfx,
			  gint x,
			  gint y)
{
    x -= spritegfx->hotx;
    y -= spritegfx->hoty;
    gsf_spritegfx_p2c(spritegfx, &x, &y);
    return gsf_collisionmask_collides_at(spritegfx->collmask, x, y);
}


void
gsf_spritegfx_p2c(GsfSpriteGfx *spritegfx,
		  gint *x,
		  gint *y)
{
    GtkObjectClass *tmpoc;

    g_return_if_fail(spritegfx != NULL
		     && GSF_IS_SPRITEGFX(spritegfx)
		     && spritegfx->collmask != NULL);

    tmpoc = GTK_OBJECT(spritegfx)->klass;
    (* GSF_SPRITEGFX_CLASS(tmpoc)->pixmap_to_coll)(x, y);
}

void
gsf_spritegfx_p2c_rect(GsfSpriteGfx *spritegfx,
		       gint *x,
		       gint *y,
		       guint *width,
		       guint *height)
{
    gint x2, y2;
    GtkObjectClass *tmpoc;

    g_return_if_fail(spritegfx != NULL
		     && GSF_IS_SPRITEGFX(spritegfx)
		     && spritegfx->collmask != NULL);

    x2 = *x + *width - 1;
    y2 = *y + *height - 1;

    tmpoc = GTK_OBJECT(spritegfx)->klass;
    (* GSF_SPRITEGFX_CLASS(tmpoc)->pixmap_to_coll)(x, y);
    (* GSF_SPRITEGFX_CLASS(tmpoc)->pixmap_to_coll)(&x2, &y2);

    *width = x2 - *x + 1;
    *height = y2 - *y + 1;
}

void
gsf_spritegfx_pastegfxrect(GsfSpriteGfx *dest,
			   GsfSpriteGfx *src,
			   gint destx,
			   gint desty,
			   GsfRectangle *srcarea,
			   gboolean maskedcopy)
{
    GdkGC *gc;
    gint srcx, srcy;
    guint width, height;
    
    g_return_if_fail(dest != NULL && GSF_IS_SPRITEGFX(dest));
    g_return_if_fail(src != NULL && GSF_IS_SPRITEGFX(src));

    destx += dest->hotx; 
    desty += dest->hoty;
    if (dest != src) {
	destx -= src->hotx;
	desty -= src->hoty;
    }
    srcx = srcarea->x + src->hotx;
    srcy = srcarea->y + src->hoty;
    width = srcarea->width;
    height = srcarea->height;
    gc = gdk_gc_new(dest->pixmap);
    if (maskedcopy) {
	gdk_gc_set_clip_mask(gc, src->mask);
	gdk_gc_set_clip_origin(gc, destx, desty);
    }
    gdk_draw_pixmap(dest->pixmap, gc, src->pixmap,
		    srcx, srcy,
		    destx, desty,
		    width, height);
    gdk_gc_unref(gc);
    gc = gdk_gc_new(dest->mask);
    if (maskedcopy) {
	gdk_gc_set_clip_mask(gc, src->mask);
	gdk_gc_set_clip_origin(gc, destx, desty);
    }
    gdk_draw_pixmap(dest->mask, gc, src->mask,
		    srcx, srcy,
		    destx, desty,
		    width, height);
    gdk_gc_unref(gc);
    
    gsf_spritegfx_p2c(src, &srcx, &srcy);
    gsf_spritegfx_p2c_rect(dest, &destx, &desty, &width, &height);
    if (maskedcopy) 
	gsf_collisionmask_or(dest->collmask, src->collmask,
			     srcx, srcy,
			     destx, desty,
			     width, height);
    else
	gsf_collisionmask_copy(dest->collmask, src->collmask,
			       srcx, srcy,
			       destx, desty,
			       width, height);

}

void
gsf_spritegfx_pastegfx(GsfSpriteGfx *dest,
		       GsfSpriteGfx *src,
		       gint destx,
		       gint desty)
{
    GsfRectangle area;

    area.x = -src->hotx;
    area.y = -src->hoty;
    area.width = src->width;
    area.height = src->height;

    gsf_spritegfx_pastegfxrect(dest, src, destx, desty, &area, TRUE);
}


void
gsf_spritegfx_clearrect(GsfSpriteGfx *dest,
			GsfRectangle *area)
{
    GdkGC *gc;
    gint x, y;
    guint width, height;
    GdkColor trans = {pixel:0, red:0, green:0, blue:0};
    
    g_return_if_fail(dest != NULL && GSF_IS_SPRITEGFX(dest));

    x = area->x + dest->hotx;
    y = area->y + dest->hoty;
    width = area->width;
    height = area->height;
    gc = gdk_gc_new(dest->pixmap);
    gdk_gc_set_foreground(gc, &trans);
    gdk_draw_rectangle(dest->pixmap, gc, TRUE,
		       x, y,
		       width, height);
    gdk_gc_unref(gc);
    gc = gdk_gc_new(dest->mask);
    gdk_gc_set_foreground(gc, &trans);
    gdk_draw_rectangle(dest->mask, gc, TRUE,
		       x, y,
		       width, height);
    gdk_gc_unref(gc);
    
    gsf_spritegfx_p2c_rect(dest, &x, &y, &width, &height);
    gsf_collisionmask_clearbit(dest->collmask,
			       x, y, width, height);

}


static guint gsf_spritegfx_collmul = 1;
static guint gsf_spritegfx_colldiv = 1;

static void
gsf_spritegfx_p2c_nop(gint *x, gint *y)
{
}

static void
gsf_spritegfx_p2c_div(gint *x, gint *y)
{
    *x /= gsf_spritegfx_colldiv;
    *y /= gsf_spritegfx_colldiv;
}

static void
gsf_spritegfx_p2c_mul(gint *x, gint *y)
{
    *x *= gsf_spritegfx_collmul;
    *y *= gsf_spritegfx_collmul;
}

static void
gsf_spritegfx_p2c_muldiv(gint *x, gint *y)
{
    *x = *x * gsf_spritegfx_collmul / gsf_spritegfx_colldiv;
    *y = *y * gsf_spritegfx_collmul / gsf_spritegfx_colldiv;
}

void
gsf_spritegfx_set_collconvert(guint mul, guint div)
{
    GtkObjectClass *klass;
    void (*p2c)(gint *x, gint *y);
    
    gsf_spritegfx_collmul = MAX(1, mul);
    gsf_spritegfx_colldiv = MAX(1, div);

    p2c = (mul > 1
	   ? div > 1
	   ? gsf_spritegfx_p2c_muldiv
	   : gsf_spritegfx_p2c_mul
	   : div > 1
	   ? gsf_spritegfx_p2c_div
	   : gsf_spritegfx_p2c_nop);

    klass = gtk_type_class(gsf_spritegfx_get_type());
    GSF_SPRITEGFX_CLASS(klass)->pixmap_to_coll = p2c;
}

void
gsf_spritegfx_get_collconvert(guint *mul, guint *div)
{
    if (*mul)
	*mul = gsf_spritegfx_collmul;
    if (*div)
	*div = gsf_spritegfx_colldiv;
}

void
gsf_spritegfx_scale_collmask(GsfSpriteGfx *gfx)
{
    if (gsf_spritegfx_colldiv > 1 || gsf_spritegfx_collmul > 1)
	gsf_collisionmask_scale(gfx->collmask,
				gfx->collmask,
				gsf_spritegfx_collmul,
				gsf_spritegfx_colldiv);
}
